from supervisely.nn.benchmark.comparison.detection_visualization.vis_metrics.vis_metric import (
    BaseVisMetric,
)
from supervisely.nn.benchmark.visualization.widgets import ChartWidget, MarkdownWidget


class AveragePrecisionByClass(BaseVisMetric):
    MARKDOWN_CLASS_AP = "markdown_class_ap_polar"
    MARKDOWN_CLASS_AP_BAR = "markdown_class_ap_bar"

    def get_figure(self):
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()
        labels = dict(r="Average Precision", theta="Class")
        cls_cnt = len(self.eval_results[0].mp.cat_names)
        for i, eval_result in enumerate(self.eval_results, 1):
            # AP per-class
            ap_per_class = eval_result.mp.coco_precision[:, :, :, 0, 2].mean(axis=(0, 1))
            ap_per_class[ap_per_class == -1] = 0  # -1 is a placeholder for no GT

            trace_name = f"[{i}] {eval_result.name}"

            if cls_cnt >= 5:
                fig.add_trace(
                    go.Scatterpolar(
                        r=ap_per_class,
                        theta=eval_result.mp.cat_names,
                        name=trace_name,
                        marker=dict(color=eval_result.color),
                        hovertemplate=trace_name
                        + "<br>"
                        + labels["theta"]
                        + ": %{theta}<br>"
                        + labels["r"]
                        + ": %{r:.2f}<br>"
                        + "<extra></extra>",
                    )
                )
            else:
                fig.add_trace(
                    go.Bar(
                        x=eval_result.mp.cat_names,
                        y=ap_per_class,
                        name=trace_name,
                        width=0.2,
                        marker=dict(color=eval_result.color),
                    )
                )

        if cls_cnt >= 5:
            fig.update_layout(
                width=800,
                height=800,
                margin=dict(l=80, r=80, t=0, b=0),
                modebar_add=["resetScale"],
                showlegend=True,
                polar=dict(radialaxis_range=[0, 1]),
            )
        else:
            fig.update_layout(
                xaxis_title="Class",
                yaxis_title="Average Precision",
                yaxis=dict(range=[0, 1.1]),
                barmode="group",
                width=700,
            )

        return fig

    @property
    def markdown_widget(self) -> MarkdownWidget:
        template_name = self.MARKDOWN_CLASS_AP
        if len(self.eval_results[0].mp.cat_names) < 5:
            template_name = self.MARKDOWN_CLASS_AP_BAR
        text: str = getattr(self.vis_texts, template_name).format(
            self.vis_texts.definitions.average_precision
        )
        return MarkdownWidget(
            name=self.MARKDOWN_CLASS_AP, title="Average Precision by Class", text=text
        )

    @property
    def chart_widget(self) -> ChartWidget:
        chart = ChartWidget(name="chart_class_ap", figure=self.get_figure())
        chart.set_click_data(
            gallery_id=self.explore_modal_table.id,
            click_data=self.get_click_data(),
            chart_click_extra="'getKey': (payload) => `${payload.points[0].curveNumber}${'_'}${payload.points[0].theta}`,",
        )
        return chart

    def get_click_data(self):
        res = {}
        res["layoutTemplate"] = [None, None, None]
        res["clickData"] = {}

        for i, eval_result in enumerate(self.eval_results):
            model_name = f"Model {i}"
            for cat_name, v in eval_result.click_data.objects_by_class.items():
                key = f"{i}_{cat_name}"
                ap_per_class_dict = res["clickData"].setdefault(key, {})

                img_ids = set()
                obj_ids = set()

                title = f"{model_name}, class: {len(v)} object{'s' if len(v) > 1 else ''}"
                ap_per_class_dict["title"] = title

                for x in v:
                    img_ids.add(x["dt_img_id"])
                    obj_ids.add(x["dt_obj_id"])

                ap_per_class_dict["imagesIds"] = list(img_ids)
                ap_per_class_dict["filters"] = [
                    {
                        "type": "tag",
                        "tagId": "confidence",
                        "value": [eval_result.f1_optimal_conf, 1],
                    },
                    {"type": "tag", "tagId": "outcome", "value": "TP"},
                    {"type": "specific_objects", "tagId": None, "value": list(obj_ids)},
                ]

        return res



from __future__ import annotations
import copy
from typing import List, Literal, Optional
from supervisely.app.widgets import Widget
from supervisely.app.content import StateJson

from supervisely.app.widgets.nodes_flow.option_components import (
    OptionComponent,
    HtmlOptionComponent,
    WidgetOptionComponent,
    ButtonOptionComponent,
    CheckboxOptionComponent,
    InputOptionComponent,
    IntegerOptionComponent,
    NumberOptionComponent,
    SelectOptionComponent,
    SliderOptionComponent,
    TextOptionComponent,
    SidebarNodeInfoOptionComponent,
)


class NodesFlow(Widget):
    class OptionComponent(OptionComponent):
        pass

    class HtmlOptionComponent(HtmlOptionComponent):
        pass

    class WidgetOptionComponent(WidgetOptionComponent):
        pass

    class ButtonOptionComponent(ButtonOptionComponent):
        pass

    class CheckboxOptionComponent(CheckboxOptionComponent):
        pass

    class InputOptionComponent(InputOptionComponent):
        pass

    class IntegerOptionComponent(IntegerOptionComponent):
        pass

    class NumberOptionComponent(NumberOptionComponent):
        pass

    class SelectOptionComponent(SelectOptionComponent):
        pass

    class SliderOptionComponent(SliderOptionComponent):
        pass

    class TextOptionComponent(TextOptionComponent):
        pass

    class SidebarNodeInfoOptionComponent(SidebarNodeInfoOptionComponent):
        pass

    class Node:
        class Input:
            def __init__(self, name, label: Optional[str] = None, color: Optional[str] = None):
                self.name = name
                self.label = label
                self.color = color

            def to_json(self):
                j = {"name": self.name}
                if self.label is not None:
                    j.setdefault("options", {})["displayName"] = self.label
                if self.color is not None:
                    j.setdefault("options", {})["type"] = self.color
                return j

        class Output(Input):
            pass

        class Option:
            def __init__(self, name: str, option_component: OptionComponent):
                if isinstance(option_component, NodesFlow.SidebarNodeInfoOptionComponent):
                    name = "sidebarNodeInfo"
                self.name = name
                self.option_component = option_component

            def to_json(self):
                return {"name": self.name, **self.option_component.to_json()}

        def __init__(
            self,
            id,
            name="Node",
            width: Optional[int] = 200,
            options: List[Option] = [],
            inputs: List[Input] = [],
            outputs: List[Output] = [],
            inputs_up: bool = False,
            position: Optional[dict] = None,
            header_color: Optional[str] = None,
            header_text_color: Optional[str] = None,
            icon: Optional[str] = None,
            icon_background_color: Optional[str] = None,
        ):
            self.id = id
            self.name = name
            self._width = width
            self._inputs_up = inputs_up
            self.options = options
            self.inputs = inputs
            self.outputs = outputs
            self._position = position
            self._header_background_color = header_color
            self._header_color = header_text_color
            self._icon = icon
            self._icon_background_color = icon_background_color

        def to_json(self):
            return {
                "id": self.id,
                "name": self.name,
                "width": self._width,
                "twoColumn": self._inputs_up,
                "options": [option.to_json() for option in self.options],
                "inputs": [i.to_json() for i in self.inputs],
                "outputs": [o.to_json() for o in self.outputs],
                "position": self._position,
                "header": {
                    "backgroundColor": self._header_background_color,
                    "color": self._header_color,
                    "icon": {
                        "backgroundColor": self._icon_background_color,
                        "class": self._icon,
                    },
                },
            }

        def set_position(self, position):
            self._position = position

    class Routes:
        SAVE = "save_cb"
        FLOW_CHANGED = "flow_changed_cb"
        FLOW_STATE_CHANGED = "flow_state_changed_cb"
        CONTEXT_MENU_CLICKED = "context_menu_item_click_cb"
        SIDEBAR_TOGGLED = "sidebar_toggled_cb"
        ITEM_DROPPED = "item_dropped_cb"
        NODE_REMOVED = "node_removed_cb"

    def __init__(
        self,
        nodes: List[Node] = [],
        height: str = None,
        context_menu: dict = None,
        color_theme: Literal["light", "dark"] = "light",
        drag_and_drop_menu: dict = None,
        drag_and_drop_menu_width: int = "400px",
        show_save: bool = True,
        widget_id: str = None,
    ):
        self._nodes = nodes
        self._height = height if height is not None else "500px"
        self._context_menu = context_menu
        self._color_theme = color_theme
        self._dd_menu = drag_and_drop_menu
        self._show_dd_area = False
        self._dd_section_width = drag_and_drop_menu_width
        self._show_save = show_save
        if self._dd_menu:
            self._show_dd_area = True

        self._save_handled = False
        self._flow_change_handled = False
        self._flow_state_change_handled = False
        self._contex_menu_item_click_handled = False
        self._sidebar_toggle_handled = False
        self._item_dropped_handled = False
        self._node_removed_handled = False

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "height": self._height,
            "contextMenuItems": self._context_menu,
            "colorTheme": self._color_theme,
            "nodeTypeList": self._dd_menu,
            "showDDArea": self._show_dd_area,
            "ddSectionWidth": self._dd_section_width,
            "interfaceTypes": [{"name": f"{color}", "color": color} for color in []],
            "showSave": self._show_save,
        }

    def get_json_state(self):
        return {
            "flowState": {},
            "flow": {"nodes": [node.to_json() for node in self._nodes], "edges": []},
        }

    def add_node(self, node: Node):
        self._nodes.append(node)
        StateJson()[self.widget_id]["flow"]["nodes"].append(node.to_json())
        StateJson().send_changes()

    def pop_node(self, idx: int):
        try:
            removed = StateJson()[self.widget_id]["flow"]["nodes"].pop(idx)
            StateJson().send_changes()
        except IndexError:
            return None
        return removed

    def delete_node_by_id(self, id: str):
        for i, node in enumerate(self._nodes):
            if node.id == id:
                return self.pop_node(i)
        return None

    def get_flow_json(self):
        return copy.deepcopy(StateJson()[self.widget_id]["flow"])

    def get_nodes_json(self):
        return copy.deepcopy(StateJson()[self.widget_id]["flow"]["nodes"])

    def get_edges_json(self):
        return copy.deepcopy(StateJson()[self.widget_id]["flow"]["edges"])

    def get_flow_state(self):
        return copy.deepcopy(StateJson()[self.widget_id]["flowState"])

    def get_nodes_state_json(self):
        """Alias for get_flow_state"""
        return self.get_flow_state()

    def on_save(self, func):
        route_path = self.get_route_path(NodesFlow.Routes.SAVE)
        server = self._sly_app.get_server()
        self._save_handled = True

        @server.post(route_path)
        def _click():
            func()

        return _click

    def flow_changed(self, func):
        route_path = self.get_route_path(NodesFlow.Routes.FLOW_CHANGED)
        server = self._sly_app.get_server()
        self._flow_change_handled = True

        @server.post(route_path)
        def _click():
            func()

        return _click

    def flow_state_changed(self, func):
        route_path = self.get_route_path(NodesFlow.Routes.FLOW_STATE_CHANGED)
        server = self._sly_app.get_server()
        self._flow_state_change_handled = True

        @server.post(route_path)
        def _click():
            func()

        return _click

    def clear(self):
        StateJson()[self.widget_id]["flow"]["nodes"] = []
        StateJson()[self.widget_id]["flow"]["edges"] = []
        StateJson()[self.widget_id]["flowState"] = {}

        self._nodes = []
        StateJson().send_changes()

    def set_edges(self, edges: List[dict]):
        StateJson()[self.widget_id]["flow"]["edges"] = edges
        StateJson().send_changes()

    def context_menu_clicked(self, func):
        route_path = self.get_route_path(NodesFlow.Routes.CONTEXT_MENU_CLICKED)
        server = self._sly_app.get_server()
        self._contex_menu_item_click_handled = True

        @server.post(route_path)
        def _click():
            item = StateJson()[self.widget_id]["selectedContextItem"]
            func(item)

        return _click

    def sidebar_toggled(self, func):
        route_path = self.get_route_path(NodesFlow.Routes.SIDEBAR_TOGGLED)
        server = self._sly_app.get_server()
        self._sidebar_toggle_handled = True

        @server.post(route_path)
        def _click():
            func()

        return _click

    def item_dropped(self, func):
        route_path = self.get_route_path(NodesFlow.Routes.ITEM_DROPPED)
        server = self._sly_app.get_server()
        self._item_dropped_handled = True

        @server.post(route_path)
        def _click():
            item = StateJson()[self.widget_id]["droppedItem"]
            func(item)

        return _click

    def node_removed(self, func):
        route_path = self.get_route_path(NodesFlow.Routes.NODE_REMOVED)
        server = self._sly_app.get_server()
        self._node_removed_handled = True

        @server.post(route_path)
        def _click():
            item = StateJson()[self.widget_id]["removedNode"]
            func(item)

        return _click



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from supervisely.nn.benchmark.visualization.vis_metric_base import MetricVis
from supervisely.nn.benchmark.visualization.vis_widgets import Schema, Widget

if TYPE_CHECKING:
    from supervisely.nn.benchmark.visualization.visualizer import Visualizer


class OutcomeCounts(MetricVis):

    def __init__(self, loader: Visualizer) -> None:
        super().__init__(loader)

        self.clickable: bool = True
        self.schema = Schema(
            self._loader.vis_texts,
            markdown_outcome_counts=Widget.Markdown(
                title="Outcome Counts",
                is_header=True,
                formats=[
                    self._loader.vis_texts.definitions.true_positives,
                    self._loader.vis_texts.definitions.false_positives,
                    self._loader.vis_texts.definitions.false_negatives,
                ],
            ),
            chart=Widget.Chart(),
        )

    def get_figure(self, widget: Widget.Chart):  # -> Optional[go.Figure]:
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()
        fig.add_trace(
            go.Bar(
                x=[self._loader.mp.TP_count],
                y=["Outcome"],
                name="TP",
                orientation="h",
                marker=dict(color="#8ACAA1"),
                hovertemplate="TP: %{x} objects<extra></extra>",
            )
        )
        fig.add_trace(
            go.Bar(
                x=[self._loader.mp.FN_count],
                y=["Outcome"],
                name="FN",
                orientation="h",
                marker=dict(color="#dd3f3f"),
                hovertemplate="FN: %{x} objects<extra></extra>",
            )
        )
        fig.add_trace(
            go.Bar(
                x=[self._loader.mp.FP_count],
                y=["Outcome"],
                name="FP",
                orientation="h",
                marker=dict(color="#F7ADAA"),
                hovertemplate="FP: %{x} objects<extra></extra>",
            )
        )
        fig.update_layout(
            barmode="stack",
            width=600,
            height=300,
        )
        fig.update_xaxes(title_text="Count (objects)")
        fig.update_yaxes(tickangle=-90)

        fig.update_layout(
            dragmode=False,
            modebar=dict(
                remove=[
                    "zoom2d",
                    "pan2d",
                    "select2d",
                    "lasso2d",
                    "zoomIn2d",
                    "zoomOut2d",
                    "autoScale2d",
                    "resetScale2d",
                ]
            ),
        )
        return fig

    def get_click_data(self, widget: Widget.Chart) -> Optional[dict]:
        if not self.clickable:
            return
        res = {}

        res["layoutTemplate"] = [None, None, None]
        res["clickData"] = {}
        for outcome, matches_data in self._loader.click_data.outcome_counts.items():
            res["clickData"][outcome] = {}
            res["clickData"][outcome]["imagesIds"] = []

            img_ids = set()
            for match_data in matches_data:
                img_comparison_data = self._loader.comparison_data[match_data["gt_img_id"]]
                if outcome == "FN":
                    img_ids.add(img_comparison_data.diff_image_info.id)
                else:
                    img_ids.add(img_comparison_data.pred_image_info.id)

            res["clickData"][outcome][
                "title"
            ] = f"{outcome}: {len(matches_data)} object{'s' if len(matches_data) > 1 else ''}"
            res["clickData"][outcome]["imagesIds"] = list(img_ids)
            res["clickData"][outcome]["filters"] = [
                {"type": "tag", "tagId": "confidence", "value": [0, 1]},
                {"type": "tag", "tagId": "outcome", "value": outcome},
            ]

        return res



from __future__ import annotations

from typing import TYPE_CHECKING

import pandas as pd

from supervisely.nn.benchmark.visualization.vis_metric_base import MetricVis
from supervisely.nn.benchmark.visualization.vis_widgets import Schema, Widget

if TYPE_CHECKING:
    from supervisely.nn.benchmark.visualization.visualizer import Visualizer


class PRCurveByClass(MetricVis):

    def __init__(self, loader: Visualizer) -> None:
        super().__init__(loader)
        self.clickable = True
        self.schema = Schema(
            self._loader.vis_texts,
            markdown_pr_by_class=Widget.Markdown(title="PR Curve by Class"),
            chart=Widget.Chart(),
        )

    def get_figure(self, widget: Widget.Chart):  # -> Optional[go.Figure]:
        import plotly.express as px  # pylint: disable=import-error

        # Precision-Recall curve per-class
        df = pd.DataFrame(self._loader.mp.pr_curve(), columns=self._loader.mp.cat_names)

        fig = px.line(
            df,
            x=self._loader.mp.recThrs,
            y=df.columns,
            # title="Precision-Recall Curve per Class",
            labels={"x": "Recall", "value": "Precision", "variable": "Category"},
            color_discrete_sequence=px.colors.qualitative.Prism,
            width=800,
            height=600,
        )

        fig.update_yaxes(range=[0, 1])
        fig.update_xaxes(range=[0, 1])
        # fig.show()

        return fig



# coding: utf-8
"""download/upload/edit :class:`Annotation<supervisely.annotation.annotation.Annotation>`"""

# docs
from __future__ import annotations

import json
from collections import defaultdict
from copy import deepcopy
from typing import Any, Callable, Dict, List, NamedTuple, Optional, Union

from tqdm import tqdm

from supervisely._utils import batched
from supervisely.annotation.annotation import Annotation, AnnotationJsonFields
from supervisely.annotation.label import Label, LabelJsonFields
from supervisely.api.module_api import ApiField, ModuleApi
from supervisely.geometry.alpha_mask import AlphaMask
from supervisely.geometry.constants import BITMAP
from supervisely.project.project_meta import ProjectMeta


class AnnotationInfo(NamedTuple):
    """
    AnnotationInfo
    """

    image_id: int
    image_name: str
    annotation: dict
    created_at: str
    updated_at: str

    def to_json(self) -> Dict[str, Any]:
        """
        Convert AnnotationInfo to JSON format.

        :return: AnnotationInfo in JSON format.
        :rtype: :class:`Dict[str, Any]`
        """
        return {
            ApiField.IMAGE_ID: self.image_id,
            ApiField.IMAGE_NAME: self.image_name,
            ApiField.ANNOTATION: self.annotation,
            ApiField.CREATED_AT: self.created_at,
            ApiField.UPDATED_AT: self.updated_at,
        }


class AnnotationApi(ModuleApi):
    """
    Annotation for a single image. :class:`AnnotationApi<AnnotationApi>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervise.ly", token="4r47N...xaTatb")

        dataset_id = 254737
        ann_infos = api.annotation.get_list(dataset_id)
    """

    @staticmethod
    def info_sequence():
        """
        NamedTuple AnnotationInfo information about Annotation.

        :Example:

         .. code-block:: python

            AnnotationInfo(image_id=121236919,
                           image_name='IMG_1836',
                           annotation={'description': '', 'tags': [], 'size': {'height': 800, 'width': 1067}, 'objects': []},
                           created_at='2019-12-19T12:06:59.435Z',
                           updated_at='2021-02-06T11:07:26.080Z')
        """
        return [
            ApiField.IMAGE_ID,
            ApiField.IMAGE_NAME,
            ApiField.ANNOTATION,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
        ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **AnnotationInfo**.
        """
        return "AnnotationInfo"

    def get_list(
        self,
        dataset_id: int,
        filters: Optional[List[Dict[str, str]]] = None,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        force_metadata_for_links: Optional[bool] = True,
    ) -> List[AnnotationInfo]:
        """
        Get list of information about all annotations for a given dataset.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param filters: List of parameters to sort output Annotations.
        :type filters: List[dict], optional
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: Information about Annotations. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[AnnotationInfo]`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dataset_id = 254737
            ann_infos = api.annotation.get_list(dataset_id)
            print(json.dumps(ann_infos[0], indent=4))
            # Output: [
            #     121236918,
            #     "IMG_0748.jpeg",
            #     {
            #         "description": "",
            #         "tags": [],
            #         "size": {
            #             "height": 800,
            #             "width": 1067
            #         },
            #         "objects": []
            #     },
            #     "2019-12-19T12:06:59.435Z",
            #     "2021-02-06T11:07:26.080Z"
            # ]

            ann_infos_filter = api.annotation.get_list(dataset_id, filters={ 'field': 'name', 'operator': '=', 'value': 'IMG_1836' })
            print(json.dumps(ann_infos_filter, indent=4))
            # Output: [
            #     121236919,
            #     "IMG_1836",
            #     {
            #         "description": "",
            #         "tags": [],
            #         "size": {
            #             "height": 800,
            #             "width": 1067
            #         },
            #         "objects": []
            #     },
            #     "2019-12-19T12:06:59.435Z",
            #     "2021-02-06T11:07:26.080Z"
            # ]
        """
        return self.get_list_all_pages(
            "annotations.list",
            {
                ApiField.DATASET_ID: dataset_id,
                ApiField.FILTER: filters or [],
                ApiField.FORCE_METADATA_FOR_LINKS: force_metadata_for_links,
            },
            progress_cb,
        )

    def get_list_generator(
        self,
        dataset_id: int,
        filters: Optional[List[Dict[str, str]]] = None,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        batch_size: Optional[int] = 50,
        force_metadata_for_links: Optional[bool] = True,
    ) -> List[AnnotationInfo]:
        """
        Get list of information about all annotations for a given dataset.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param filters: List of parameters to sort output Annotations.
        :type filters: List[dict], optional
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: Information about Annotations. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[AnnotationInfo]`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dataset_id = 254737
            ann_infos = api.annotation.get_list(dataset_id)
            print(json.dumps(ann_infos[0], indent=4))
            # Output: [
            #     121236918,
            #     "IMG_0748.jpeg",
            #     {
            #         "description": "",
            #         "tags": [],
            #         "size": {
            #             "height": 800,
            #             "width": 1067
            #         },
            #         "objects": []
            #     },
            #     "2019-12-19T12:06:59.435Z",
            #     "2021-02-06T11:07:26.080Z"
            # ]

            ann_infos_filter = api.annotation.get_list(dataset_id, filters={ 'field': 'name', 'operator': '=', 'value': 'IMG_1836' })
            print(json.dumps(ann_infos_filter, indent=4))
            # Output: [
            #     121236919,
            #     "IMG_1836",
            #     {
            #         "description": "",
            #         "tags": [],
            #         "size": {
            #             "height": 800,
            #             "width": 1067
            #         },
            #         "objects": []
            #     },
            #     "2019-12-19T12:06:59.435Z",
            #     "2021-02-06T11:07:26.080Z"
            # ]
        """
        data = {
            ApiField.DATASET_ID: dataset_id,
            ApiField.FILTER: filters or [],
            ApiField.FORCE_METADATA_FOR_LINKS: force_metadata_for_links,
            ApiField.PAGINATION_MODE: ApiField.TOKEN,
        }
        if batch_size is not None:
            data[ApiField.PER_PAGE] = batch_size
        else:
            # use default value on instance (learn in API documentation)
            # 20k for instance
            # 50 by default in SDK
            pass

        return self.get_list_all_pages_generator("annotations.list", data, progress_cb)

    def download(
        self,
        image_id: int,
        with_custom_data: Optional[bool] = False,
        force_metadata_for_links: Optional[bool] = True,
    ) -> AnnotationInfo:
        """
        Download AnnotationInfo by image ID from API.

        :param image_id: Image ID in Supervisely.
        :type image_id: int
        :param with_custom_data: Include custom data in the response.
        :type with_custom_data: bool, optional
        :param force_metadata_for_links: Force metadata for links.
        :type force_metadata_for_links: bool, optional

        :return: Information about Annotation. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`AnnotationInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            image_id = 121236918
            ann_info = api.annotation.download(image_id)
            print(json.dumps(ann_info, indent=4))
            # Output: [
            #     121236918,
            #     "IMG_0748.jpeg",
            #     {
            #         "description": "",
            #         "tags": [],
            #         "size": {
            #             "height": 800,
            #             "width": 1067
            #         },
            #         "objects": []
            #     },
            #     "2019-12-19T12:06:59.435Z",
            #     "2021-02-06T11:07:26.080Z"
            # ]
        """
        response = self._api.post(
            "annotations.info",
            {
                ApiField.IMAGE_ID: image_id,
                ApiField.WITH_CUSTOM_DATA: with_custom_data,
                ApiField.FORCE_METADATA_FOR_LINKS: force_metadata_for_links,
                ApiField.INTEGER_COORDS: False,
            },
        )
        result = response.json()
        # Convert annotation to pixel coordinate system
        result[ApiField.ANNOTATION] = Annotation._to_pixel_coordinate_system_json(
            result[ApiField.ANNOTATION]
        )
        # check if there are any AlphaMask geometries in the batch
        additonal_geometries = defaultdict(int)
        labels = result[ApiField.ANNOTATION][AnnotationJsonFields.LABELS]
        for idx, label in enumerate(labels):
            if label[LabelJsonFields.GEOMETRY_TYPE] == AlphaMask.geometry_name():
                figure_id = label[LabelJsonFields.ID]
                additonal_geometries[figure_id] = idx

        # if so, download them separately and update the annotation
        if len(additonal_geometries) > 0:
            figure_ids = list(additonal_geometries.keys())
            figures = self._api.image.figure.download_geometries_batch(figure_ids)
            for figure_id, geometry in zip(figure_ids, figures):
                label_idx = additonal_geometries[figure_id]
                labels[label_idx].update({BITMAP: geometry})
        ann_info = self._convert_json_info(result)

        return ann_info

    def download_json(
        self,
        image_id: int,
        with_custom_data: Optional[bool] = False,
        force_metadata_for_links: Optional[bool] = True,
    ) -> Dict[str, Union[str, int, list, dict]]:
        """
        Download Annotation in json format by image ID from API.

        :param image_id: Image ID in Supervisely.
        :type image_id: int
        :param with_custom_data: Include custom data in the response.
        :type with_custom_data: bool, optional
        :param force_metadata_for_links: Force metadata for links.
        :type force_metadata_for_links: bool, optional

        :return: Annotation in json format
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            image_id = 121236918
            ann_json = api.annotation.download_json(image_id)
            print(ann_json)
            # Output: {
            #         "description": "",
            #         "tags": [],
            #         "size": {
            #             "height": 800,
            #             "width": 1067
            #         },
            #         "objects": []
            #     }
        """
        return self.download(
            image_id=image_id,
            with_custom_data=with_custom_data,
            force_metadata_for_links=force_metadata_for_links,
        ).annotation

    def download_batch(
        self,
        dataset_id: int,
        image_ids: List[int],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        with_custom_data: Optional[bool] = False,
        force_metadata_for_links: Optional[bool] = True,
    ) -> List[AnnotationInfo]:
        """
        Get list of AnnotationInfos for given dataset ID from API.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param image_ids: List of integers.
        :type image_ids: List[int]
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm
        :param with_custom_data: Include custom data in the response.
        :type with_custom_data: bool, optional
        :param force_metadata_for_links: Force metadata for links.
        :type force_metadata_for_links: bool, optional

        :return: Information about Annotations. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[AnnotationInfo]`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dataset_id = 254737
            image_ids = [121236918, 121236919]
            p = tqdm(desc="Annotations downloaded: ", total=len(image_ids))

            ann_infos = api.annotation.download_batch(dataset_id, image_ids, progress_cb=p)
            # Output:
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Annotations downloaded: ", "current": 0, "total": 2, "timestamp": "2021-03-16T15:20:06.168Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Annotations downloaded: ", "current": 2, "total": 2, "timestamp": "2021-03-16T15:20:06.510Z", "level": "info"}

            Optimizing the download process by using the context to avoid redundant API calls.:
            # 1. Download the project meta
            project_id = api.dataset.get_info_by_id(dataset_id).project_id
            project_meta = api.project.get_meta(project_id)

            # 2. Use the context to avoid redundant API calls
            dataset_id = 254737
            image_ids = [121236918, 121236919]
            with sly.ApiContext(api, dataset_id=dataset_id, project_id=project_id, project_meta=project_meta):
                ann_infos = api.annotation.download_batch(dataset_id, image_ids)
        """
        # use context to avoid redundant API calls
        context = self._api.optimization_context
        context_dataset_id = context.get("dataset_id")
        project_meta = context.get("project_meta")
        project_id = context.get("project_id")
        if dataset_id != context_dataset_id:
            context["dataset_id"] = dataset_id
            project_id, project_meta = None, None

        if not isinstance(project_meta, ProjectMeta):
            if project_id is None:
                project_id = self._api.dataset.get_info_by_id(dataset_id).project_id
                context["project_id"] = project_id
            project_meta = ProjectMeta.from_json(self._api.project.get_meta(project_id))
            context["project_meta"] = project_meta

        need_download_alpha_masks = False
        for obj_cls in project_meta.obj_classes:
            if obj_cls.geometry_type == AlphaMask:
                need_download_alpha_masks = True
                break

        id_to_ann = {}
        for batch in batched(image_ids):
            post_data = {
                ApiField.DATASET_ID: dataset_id,
                ApiField.IMAGE_IDS: batch,
                ApiField.WITH_CUSTOM_DATA: with_custom_data,
                ApiField.FORCE_METADATA_FOR_LINKS: force_metadata_for_links,
                ApiField.INTEGER_COORDS: False,
            }
            results = self._api.post("annotations.bulk.info", data=post_data).json()
            if need_download_alpha_masks is True:
                additonal_geometries = defaultdict(tuple)
                for ann_idx, ann_dict in enumerate(results):
                    # check if there are any AlphaMask geometries in the batch
                    for label_idx, label in enumerate(
                        ann_dict[ApiField.ANNOTATION][AnnotationJsonFields.LABELS]
                    ):
                        if label[LabelJsonFields.GEOMETRY_TYPE] == AlphaMask.geometry_name():
                            figure_id = label[LabelJsonFields.ID]
                            additonal_geometries[figure_id] = (ann_idx, label_idx)

                # if there are any AlphaMask geometries, download them separately and update the annotation
                if len(additonal_geometries) > 0:
                    figure_ids = list(additonal_geometries.keys())
                    figures = self._api.image.figure.download_geometries_batch(figure_ids)
                    for figure_id, geometry in zip(figure_ids, figures):
                        ann_idx, label_idx = additonal_geometries[figure_id]
                        results[ann_idx][ApiField.ANNOTATION][AnnotationJsonFields.LABELS][
                            label_idx
                        ].update({BITMAP: geometry})

            for ann_dict in results:
                # Convert annotation to pixel coordinate system
                ann_dict[ApiField.ANNOTATION] = Annotation._to_pixel_coordinate_system_json(
                    ann_dict[ApiField.ANNOTATION]
                )
                ann_info = self._convert_json_info(ann_dict)
                id_to_ann[ann_info.image_id] = ann_info

            if progress_cb is not None:
                progress_cb(len(batch))
        ordered_results = [id_to_ann[image_id] for image_id in image_ids]
        return ordered_results

    def download_json_batch(
        self,
        dataset_id: int,
        image_ids: List[int],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        force_metadata_for_links: Optional[bool] = True,
    ) -> List[Dict]:
        """
        Get list of AnnotationInfos for given dataset ID from API.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param image_ids: List of integers.
        :type image_ids: List[int]
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm
        :param force_metadata_for_links: Force metadata for links.
        :type force_metadata_for_links: bool, optional

        :return: Information about Annotations. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[Dict]`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dataset_id = 254737
            image_ids = [121236918, 121236919]
            p = tqdm(desc="Annotations downloaded: ", total=len(image_ids))

            anns_jsons = api.annotation.download_json_batch(dataset_id, image_ids, progress_cb=p)
            # Output:
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Annotations downloaded: ", "current": 0, "total": 2, "timestamp": "2021-03-16T15:20:06.168Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Annotations downloaded: ", "current": 2, "total": 2, "timestamp": "2021-03-16T15:20:06.510Z", "level": "info"}
        """
        results = self.download_batch(
            dataset_id=dataset_id,
            image_ids=image_ids,
            progress_cb=progress_cb,
            force_metadata_for_links=force_metadata_for_links,
        )
        return [ann_info.annotation for ann_info in results]

    def upload_path(
        self,
        img_id: int,
        ann_path: str,
        skip_bounds_validation: Optional[bool] = False,
    ) -> None:
        """
        Loads an annotation from a given path to a given image ID in the API.

        :param img_id: Image ID in Supervisely.
        :type img_id: int
        :param ann_path: Path to annotation on host.
        :type ann_path: str
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            image_id = 121236918
            ann_path = '/home/admin/work/supervisely/example/ann.json'
            upl_path = api.annotation.upload_path(image_id, ann_path)
        """
        self.upload_paths([img_id], [ann_path], skip_bounds_validation=skip_bounds_validation)

    def upload_paths(
        self,
        img_ids: List[int],
        ann_paths: List[str],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        skip_bounds_validation: Optional[bool] = False,
    ) -> None:
        """
        Loads an annotations from a given paths to a given images IDs in the API. Images IDs must be from one dataset.

        :param img_ids: Images IDs in Supervisely.
        :type img_ids: List[int]
        :param ann_paths: Paths to annotations on local machine.
        :type ann_paths: List[str]
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            img_ids = [121236918, 121236919]
            ann_pathes = ['/home/admin/work/supervisely/example/ann1.json', '/home/admin/work/supervisely/example/ann2.json']
            upl_paths = api.annotation.upload_paths(img_ids, ann_pathes)

            # Optimizing the upload process by using the context to avoid redundant API calls.
            # Usefull when uploading a large number of annotations in one dataset.
            # 1. Download the project meta
            dataset_id = 254737
            project_id = api.dataset.get_info_by_id(dataset_id).project_id
            project_meta = api.project.get_meta(project_id)

            # 2. Use the context to avoid redundant API calls
            with sly.ApiContext(api, dataset_id=dataset_id, project_id=project_id, project_meta=project_meta):
                api.annotation.upload_paths(img_ids, ann_pathes)
        """

        def read_json(ann_path):
            with open(ann_path) as json_file:
                return json.load(json_file)

        self._upload_batch(
            read_json,
            img_ids,
            ann_paths,
            progress_cb,
            skip_bounds_validation=skip_bounds_validation,
        )

    def upload_json(
        self,
        img_id: int,
        ann_json: Dict,
        skip_bounds_validation: Optional[bool] = False,
    ) -> None:
        """
        Loads an annotation from dict to a given image ID in the API.

        :param img_id: Image ID in Supervisely.
        :type img_id: int
        :param ann_json: Annotation in JSON format.
        :type ann_json: dict
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            image_id = 121236918
            upl_json = api.annotation.upload_json(image_id, ann_json)
        """
        self.upload_jsons([img_id], [ann_json], skip_bounds_validation=skip_bounds_validation)

    def upload_jsons(
        self,
        img_ids: List[int],
        ann_jsons: List[Dict],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        skip_bounds_validation: Optional[bool] = False,
    ) -> None:
        """
        Loads an annotations from dicts to a given images IDs in the API. Images IDs must be from one dataset.

        :param img_ids: Image ID in Supervisely.
        :type img_ids: List[int]
        :param ann_jsons: Annotation in JSON format.
        :type ann_jsons: List[dict]
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            img_ids = [121236918, 121236919]
            api.annotation.upload_jsons(img_ids, ann_jsons)

            # Optimizing the upload process by using the context to avoid redundant API calls.
            # Usefull when uploading a large number of annotations in one dataset.
            # 1. Download the project meta
            dataset_id = 254737
            project_id = api.dataset.get_info_by_id(dataset_id).project_id
            project_meta = api.project.get_meta(project_id)

            # 2. Use the context to avoid redundant API calls
            with sly.ApiContext(api, dataset_id=dataset_id, project_id=project_id, project_meta=project_meta):
                api.annotation.upload_jsons(img_ids, ann_jsons)
        """
        self._upload_batch(
            lambda x: x,
            img_ids,
            ann_jsons,
            progress_cb,
            skip_bounds_validation=skip_bounds_validation,
        )

    def upload_ann(
        self,
        img_id: int,
        ann: Annotation,
        skip_bounds_validation: Optional[bool] = False,
    ) -> None:
        """
        Loads an :class:`Annotation<supervisely.annotation.annotation.Annotation>` to a given image ID in the API.

        :param img_id: Image ID in Supervisely.
        :type img_id: int
        :param ann: Annotation object.
        :type ann: Annotation
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            image_id = 121236918
            upl_ann = api.annotation.upload_ann(image_id, ann)
        """
        self.upload_anns([img_id], [ann], skip_bounds_validation=skip_bounds_validation)

    def upload_anns(
        self,
        img_ids: List[int],
        anns: List[Annotation],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        skip_bounds_validation: Optional[bool] = False,
    ) -> None:
        """
        Loads an :class:`Annotations<supervisely.annotation.annotation.Annotation>` to a given images IDs in the API. Images IDs must be from one dataset.

        :param img_ids: Image ID in Supervisely.
        :type img_ids: List[int]
        :param anns: List of Annotation objects.
        :type anns: List[Annotation]
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            img_ids = [121236918, 121236919]
            upl_anns = api.annotation.upload_anns(img_ids, [ann1, ann2])

            # Optimizing the upload process by using the context to avoid redundant API calls.
            # Usefull when uploading a large number of annotations in one dataset.
            # 1. Download the project meta
            dataset_id = 254737
            project_id = api.dataset.get_info_by_id(dataset_id).project_id
            project_meta = api.project.get_meta(project_id)

            # 2. Use the context to avoid redundant API calls
            with sly.ApiContext(api, dataset_id=dataset_id, project_id=project_id, project_meta=project_meta):
                api.annotation.upload_anns(img_ids, [ann1, ann2])
        """
        # img_ids from the same dataset
        self._upload_batch(
            Annotation.to_json,
            img_ids,
            anns,
            progress_cb,
            skip_bounds_validation=skip_bounds_validation,
        )

    def _upload_batch(
        self,
        func_ann_to_json: Callable,
        img_ids: List[int],
        anns: List[Union[Dict, Annotation, str]],
        progress_cb=None,
        skip_bounds_validation: Optional[bool] = False,
    ):
        """
        General method for uploading annotations to instance.

        Method is used in: upload_paths, upload_jsons, upload_anns

        :param func_ann_to_json: Function to convert annotation to json or read annotation from file.
        :type func_ann_to_json: callable
        :param img_ids: List of image IDs in Supervisely to which annotations will be uploaded.
        :type img_ids: List[int]
        :param anns: List of annotations. Can be json, Annotation object or path to annotation file.
        :type anns: List[Union[Dict, Annotation, str]]
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :param skip_bounds_validation: Skip bounds validation.
        :type skip_bounds_validation: bool, optional
        :return: None
        :rtype: :class:`NoneType`
        """
        # img_ids from the same dataset
        if len(img_ids) == 0:
            return
        if len(img_ids) != len(anns):
            raise RuntimeError(
                'Can not match "img_ids" and "anns" lists, len(img_ids) != len(anns)'
            )

        # use context to avoid redundant API calls
        dataset_id = self._api.image.get_info_by_id(
            img_ids[0], force_metadata_for_links=False
        ).dataset_id
        context = self._api.optimization_context
        context_dataset_id = context.get("dataset_id")
        project_id = context.get("project_id")
        project_meta = context.get("project_meta")
        if dataset_id != context_dataset_id:
            context["dataset_id"] = dataset_id
            project_id, project_meta = None, None

        if not isinstance(project_meta, ProjectMeta):
            if project_id is None:
                project_id = self._api.dataset.get_info_by_id(dataset_id).project_id
                context["project_id"] = project_id
            project_meta = ProjectMeta.from_json(self._api.project.get_meta(project_id))
            context["project_meta"] = project_meta

        need_upload_alpha_masks = False
        for obj_cls in project_meta.obj_classes:
            if obj_cls.geometry_type == AlphaMask:
                need_upload_alpha_masks = True
                break

        for batch in batched(list(zip(img_ids, anns))):
            data = []
            if need_upload_alpha_masks:
                special_figures = []
                special_geometries = []
                # check if there are any AlphaMask geometries in the batch
                for img_id, ann in batch:
                    ann_json = func_ann_to_json(ann)
                    ann_json = deepcopy(ann_json)  # Avoid changing the original data

                    ann_json = Annotation._to_subpixel_coordinate_system_json(ann_json)
                    filtered_labels = []
                    if AnnotationJsonFields.LABELS not in ann_json:
                        raise RuntimeError(
                            f"Annotation JSON does not contain '{AnnotationJsonFields.LABELS}' field"
                        )
                    for label_json in ann_json[AnnotationJsonFields.LABELS]:
                        for key in [
                            LabelJsonFields.GEOMETRY_TYPE,
                            LabelJsonFields.OBJ_CLASS_NAME,
                        ]:
                            if key not in label_json:
                                raise RuntimeError(f"Label JSON does not contain '{key}' field")
                        if label_json[LabelJsonFields.GEOMETRY_TYPE] == AlphaMask.geometry_name():
                            label_json.update({ApiField.ENTITY_ID: img_id})

                            obj_cls_name = label_json.get(LabelJsonFields.OBJ_CLASS_NAME)
                            obj_cls = project_meta.get_obj_class(obj_cls_name)
                            if obj_cls is None:
                                raise RuntimeError(
                                    f"Object class '{obj_cls_name}' not found in project meta"
                                )
                            # update obj class id in label json
                            label_json[LabelJsonFields.OBJ_CLASS_ID] = obj_cls.sly_id

                            geometry = label_json.pop(
                                BITMAP
                            )  # remove alpha mask geometry from label json
                            special_geometries.append(geometry)
                            special_figures.append(label_json)
                        else:
                            filtered_labels.append(label_json)
                    if len(filtered_labels) != len(ann_json[AnnotationJsonFields.LABELS]):
                        ann_json[AnnotationJsonFields.LABELS] = filtered_labels
                    data.append({ApiField.IMAGE_ID: img_id, ApiField.ANNOTATION: ann_json})
            else:
                for img_id, ann in batch:
                    ann_json = func_ann_to_json(ann)
                    ann_json = deepcopy(ann_json)  # Avoid changing the original data
                    ann_json = Annotation._to_subpixel_coordinate_system_json(ann_json)
                    data.append({ApiField.IMAGE_ID: img_id, ApiField.ANNOTATION: ann_json})

            self._api.post(
                "annotations.bulk.add",
                data={
                    ApiField.DATASET_ID: dataset_id,
                    ApiField.ANNOTATIONS: data,
                    ApiField.SKIP_BOUNDS_VALIDATION: skip_bounds_validation,
                },
            )
            if need_upload_alpha_masks:
                if len(special_figures) > 0:
                    # 1. create figures
                    json_body = {
                        ApiField.DATASET_ID: dataset_id,
                        ApiField.FIGURES: special_figures,
                        ApiField.SKIP_BOUNDS_VALIDATION: skip_bounds_validation,
                    }
                    resp = self._api.post("figures.bulk.add", json_body)
                    added_fig_ids = [resp_obj[ApiField.ID] for resp_obj in resp.json()]

                    # 2. upload alpha mask geometries
                    self._api.image.figure.upload_geometries_batch(
                        added_fig_ids, special_geometries
                    )

            if progress_cb is not None:
                progress_cb(len(batch))

    def get_info_by_id(self, id):
        """
        get_info_by_id
        """
        raise NotImplementedError("Method is not supported")

    def get_info_by_name(self, parent_id, name):
        """
        get_info_by_name
        """
        raise NotImplementedError("Method is not supported")

    def exists(self, parent_id, name):
        """
        exists
        """
        raise NotImplementedError("Method is not supported")

    def get_free_name(self, parent_id, name):
        """
        get_free_name
        """
        raise NotImplementedError("Method is not supported")

    def _add_sort_param(self, data):
        """
        _add_sort_param
        """
        return data

    def copy_batch(
        self,
        src_image_ids: List[int],
        dst_image_ids: List[int],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        force_metadata_for_links: Optional[bool] = True,
        skip_bounds_validation: Optional[bool] = False,
    ) -> None:
        """
        Copy annotations from one images IDs to another in API.

        :param src_image_ids: Images IDs in Supervisely.
        :type src_image_ids: List[int]
        :param dst_image_ids: Unique IDs of images in API.
        :type dst_image_ids: List[int]
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :raises: :class:`RuntimeError`, if len(src_image_ids) != len(dst_image_ids)
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from tqdm import tqdm

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            src_ids = [121236918, 121236919]
            dst_ids = [547837053, 547837054]
            p = tqdm(desc="Annotations copy: ", total=len(src_ids))

            copy_anns = api.annotation.copy_batch(src_ids, dst_ids, progress_cb=p)
            # Output:
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Annotations copy: ", "current": 0, "total": 2, "timestamp": "2021-03-16T15:24:31.286Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Annotations copy: ", "current": 2, "total": 2, "timestamp": "2021-03-16T15:24:31.288Z", "level": "info"}
        """
        if len(src_image_ids) != len(dst_image_ids):
            raise RuntimeError(
                'Can not match "src_image_ids" and "dst_image_ids" lists, '
                "len(src_image_ids) != len(dst_image_ids)"
            )
        if len(src_image_ids) == 0:
            return

        src_dataset_id = self._api.image.get_info_by_id(src_image_ids[0]).dataset_id
        for cur_batch in batched(list(zip(src_image_ids, dst_image_ids))):
            src_ids_batch, dst_ids_batch = zip(*cur_batch)
            ann_infos = self.download_batch(
                src_dataset_id,
                src_ids_batch,
                force_metadata_for_links=force_metadata_for_links,
            )
            ann_jsons = [ann_info.annotation for ann_info in ann_infos]
            self.upload_jsons(
                dst_ids_batch, ann_jsons, skip_bounds_validation=skip_bounds_validation
            )
            if progress_cb is not None:
                progress_cb(len(src_ids_batch))

    def copy(
        self,
        src_image_id: int,
        dst_image_id: int,
        force_metadata_for_links: Optional[bool] = True,
        skip_bounds_validation: Optional[bool] = False,
    ) -> None:
        """
        Copy annotation from one image ID to another image ID in API.

        :param src_image_id: Image ID in Supervisely.
        :type src_image_id: int
        :param dst_image_id: Image ID in Supervisely.
        :type dst_image_id: int
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            src_id = 121236918
            dst_id = 547837053
            api.annotation.copy(src_id, dst_id)
        """
        self.copy_batch(
            [src_image_id],
            [dst_image_id],
            force_metadata_for_links=force_metadata_for_links,
            skip_bounds_validation=skip_bounds_validation,
        )

    def copy_batch_by_ids(
        self,
        src_image_ids: List[int],
        dst_image_ids: List[int],
        batch_size: Optional[int] = 50,
        save_source_date: Optional[bool] = True,
    ) -> None:
        """
        Copy annotations from one images IDs to another images IDs in API.

        :param src_image_ids: Images IDs in Supervisely.
        :type src_image_ids: List[int]
        :param dst_image_ids: Images IDs in Supervisely.
        :type dst_image_ids: List[int]
        :return: None
        :rtype: :class:`NoneType`
        :raises: :class:`RuntimeError` if len(src_image_ids) != len(dst_image_ids)

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            src_ids = [121236918, 121236919]
            dst_ids = [547837053, 547837054]
            api.annotation.copy_batch_by_ids(src_ids, dst_ids)
        """
        if len(src_image_ids) != len(dst_image_ids):
            raise RuntimeError(
                'Can not match "src_image_ids" and "dst_image_ids" lists, '
                "len(src_image_ids) != len(dst_image_ids)"
            )
        if len(src_image_ids) == 0:
            return
        for cur_batch in batched(list(zip(src_image_ids, dst_image_ids)), batch_size=batch_size):
            src_ids_batch, dst_ids_batch = zip(*cur_batch)
            self._api.post(
                "annotations.bulk.copy",
                data={
                    "srcImageIds": src_ids_batch,
                    "destImageIds": dst_ids_batch,
                    "preserveSourceDate": save_source_date,
                },
            )

    def _convert_json_info(self, info: dict, skip_missing=True) -> AnnotationInfo:
        """
        _convert_json_info
        """
        res = super()._convert_json_info(info, skip_missing=skip_missing)
        return AnnotationInfo(**res._asdict())

    def append_labels(
        self,
        image_id: int,
        labels: List[Label],
        skip_bounds_validation: Optional[bool] = False,
    ) -> None:
        """
        Append labels to image with given ID in API.

        :param image_id: Image ID to append labels.
        :type image_id: int
        :param labels: List of labels to append.
        :type labels: List[Label]
        :return: None
        :rtype: :class:`NoneType`
        """
        if len(labels) == 0:
            return

        alpha_mask_geometry = AlphaMask._impl_json_class_name()
        payload = []
        special_geometries = {}
        for idx, label in enumerate(labels):
            _label_json = label.to_json()
            if isinstance(label.geometry, AlphaMask):
                _label_json.pop(alpha_mask_geometry)  # remove alpha mask geometry from label json
                special_geometries[idx] = label.geometry.to_json()[BITMAP]
            else:
                _label_json["geometry"] = label.geometry.to_json()
            if "classId" not in _label_json:
                raise KeyError("Update project meta from server to get class id")
            payload.append(_label_json)

        added_ids = []
        for batch_jsons in batched(payload, batch_size=100):
            resp = self._api.post(
                "figures.bulk.add",
                {
                    ApiField.ENTITY_ID: image_id,
                    ApiField.FIGURES: batch_jsons,
                    ApiField.SKIP_BOUNDS_VALIDATION: skip_bounds_validation,
                },
            )
            for resp_obj in resp.json():
                figure_id = resp_obj[ApiField.ID]
                added_ids.append(figure_id)

        # upload alpha mask geometries
        if len(special_geometries) > 0:
            fidure_ids = [added_ids[idx] for idx in special_geometries.keys()]
            self._api.image.figure.upload_geometries_batch(fidure_ids, special_geometries.values())

    def get_label_by_id(
        self, label_id: int, project_meta: ProjectMeta, with_tags: Optional[bool] = True
    ) -> Label:
        """Returns Supervisely Label object by it's ID.

        :param label_id: ID of the label to get
        :type label_id: int
        :param project_meta: Supervisely ProjectMeta object
        :type project_meta: ProjectMeta
        :param with_tags: If True, tags will be added to the Label object
        :type with_tags: bool, optional
        :return: Supervisely Label object
        :rtype: Label
        :Usage example:

         .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            load_dotenv(os.path.expanduser("~/supervisely.env"))
            api = sly.Api.from_env()

            label_id = 121236918

            project_id = 254737
            project_meta = sly.ProjectMeta.from_json(api.project.get_meta(project_id))

            label = api.annotation.get_label_by_id(label_id, project_meta)
        """
        resp = self._api.get(
            "figures.info", {ApiField.ID: label_id, ApiField.DECOMPRESS_BITMAP: False}
        )
        geometry = resp.json()

        class_id = geometry.get("classId")
        geometry["classTitle"] = project_meta.get_obj_class_by_id(class_id).name
        geometry.update(geometry.get("geometry"))
        geometry["tags"] = self._get_label_tags(label_id) if with_tags else []

        return Label.from_json(geometry, project_meta)

    def _get_label_tags(self, label_id: int) -> List[Dict[str, Any]]:
        """Returns tags of the label with given ID in JSON format.

        :param label_id: ID of the label to get tags
        :type label_id: int
        :return: list of tags in JSON format
        :rtype: List[Dict[str, Any]]
        :Usage example:

         .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            load_dotenv(os.path.expanduser("~/supervisely.env"))
            api = sly.Api.from_env()

            label_id = 121236918
            tags_json = api.annotation.get_label_tags(label_id)
        """
        return self._api.get("figures.tags.list", {ApiField.ID: label_id}).json()

    def update_label(self, label_id: int, label: Label) -> None:
        """Updates label with given ID in Supervisely with new Label object.
        NOTE: This method only updates label's geometry and tags, not class title, etc.

        :param label_id: ID of the label to update
        :type label_id: int
        :param label: Supervisely Label object
        :type label: Label
        :Usage example:

         .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            load_dotenv(os.path.expanduser("~/supervisely.env"))
            api = sly.Api.from_env()

            new_label: sly.Label
            label_id = 121236918

            api.annotation.update_label(label_id, new_label)
        """
        self._api.post(
            "figures.editInfo",
            {
                ApiField.ID: label_id,
                ApiField.TAGS: [tag.to_json() for tag in label.tags],
                ApiField.GEOMETRY: label.geometry.to_json(),
            },
        )

    def update_label_priority(self, label_id: int, priority: int) -> None:
        """Updates label's priority with given ID in Supervisely.
        Priority increases with the number: a higher number indicates a higher priority.
        The higher priority means that the label will be displayed on top of the others.
        The lower priority means that the label will be displayed below the others.
        
        :param label_id: ID of the label to update
        :type label_id: int
        :param priority: New priority of the label
        :type priority: int
        
        :Usage example: 
        
            .. code-block:: python
            
            import os
            from dotenv import load_dotenv
            
            import supervisely as sly
            
            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            load_dotenv(os.path.expanduser("~/supervisely.env"))
            
            api = sly.Api.from_env()
            
            label_ids = [123, 456, 789]
            priorities = [1, 2, 3]
            
            for label_id, priority in zip(label_ids, priorities):
                api.annotation.update_label_priority(label_id, priority)
                
            # The label with ID 789 will be displayed on top of the others.
            # The label with ID 123 will be displayed below the others.
        
        """
        self._api.post(
            "figures.priority.update",
            {
                ApiField.ID: label_id,
                ApiField.PRIORITY: priority,
            },
        )


from typing import List, Union

from supervisely.imaging.color import hex2rgb
from supervisely.nn.benchmark.comparison.detection_visualization.vis_metrics.vis_metric import (
    BaseVisMetric,
)
from supervisely.nn.benchmark.visualization.widgets import (
    ChartWidget,
    MarkdownWidget,
    TableWidget,
)


class Speedtest(BaseVisMetric):

    def is_empty(self) -> bool:
        return not any(eval_result.speedtest_info for eval_result in self.eval_results)

    @property
    def latency(self) -> List[Union[int, str]]:
        latency = []
        for eval_result in self.eval_results:
            if eval_result.speedtest_info is None:
                latency.append("N/A")
            else:
                added = False
                for test in eval_result.speedtest_info["speedtest"]:
                    if test["batch_size"] == 1:
                        latency.append(round(test["benchmark"]["total"], 2))
                        added = True
                        break
                if not added:
                    latency.append("N/A")
        return latency

    @property
    def fps(self) -> List[Union[int, str]]:
        fps = []
        for eval_result in self.eval_results:
            if eval_result.speedtest_info is None:
                fps.append("N/A")
            else:
                added = False
                for test in eval_result.speedtest_info["speedtest"]:
                    if test["batch_size"] == 1:
                        fps.append(round(1000 / test["benchmark"]["total"], 2))
                        added = True
                        break
                if not added:
                    fps.append("N/A")
        return fps

    @property
    def md_intro(self) -> MarkdownWidget:
        return MarkdownWidget(
            name="speedtest_intro",
            title="Inference Speed",
            text=self.vis_texts.markdown_speedtest_intro,
        )

    @property
    def intro_table(self) -> TableWidget:
        columns = ["Model", "Device", "Hardware", "Runtime"]
        columns_options = [{"disableSort": True} for _ in columns]
        content = []
        for i, eval_result in enumerate(self.eval_results, 1):
            name = f"[{i}] {eval_result.name}"
            if eval_result.speedtest_info is None:
                row = [name, "N/A", "N/A", "N/A"]
                dct = {
                    "row": row,
                    "id": name,
                    "items": row,
                }
                content.append(dct)
                continue
            model_info = eval_result.speedtest_info.get("model_info", {})
            device = model_info.get("device", "N/A")
            hardware = model_info.get("hardware", "N/A")
            runtime = model_info.get("runtime", "N/A")
            row = [name, device, hardware, runtime]
            dct = {
                "row": row,
                "id": name,
                "items": row,
            }
            content.append(dct)

        data = {
            "columns": columns,
            "columnsOptions": columns_options,
            "content": content,
        }
        return TableWidget(
            name="speedtest_intro_table",
            data=data,
            show_header_controls=False,
            fix_columns=1,
        )

    @property
    def inference_time_md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_speedtest_overview_ms.format(100)
        return MarkdownWidget(
            name="inference_time_md",
            title="Overview",
            text=text,
        )

    @property
    def fps_md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_speedtest_overview_fps.format(100)
        return MarkdownWidget(
            name="fps_md",
            title="FPS Table",
            text=text,
        )

    @property
    def fps_table(self) -> TableWidget:
        data = {}
        batch_sizes = set()
        max_fps = 0
        for i, eval_result in enumerate(self.eval_results, 1):
            data[i] = {}
            if eval_result.speedtest_info is None:
                continue
            speedtests = eval_result.speedtest_info["speedtest"]
            for test in speedtests:
                batch_size = test["batch_size"]
                fps = round(1000 / test["benchmark"]["total"] * batch_size)
                batch_sizes.add(batch_size)
                max_fps = max(max_fps, fps)
                data[i][batch_size] = fps

        batch_sizes = sorted(batch_sizes)
        columns = ["Model"]
        columns_options = [{"disableSort": True}]
        for batch_size in batch_sizes:
            columns.append(f"Batch size {batch_size}")
            columns_options.append(
                {
                    "subtitle": "imgs/sec",
                    "tooltip": "Frames (images) per second",
                    "postfix": "fps",
                    # "maxValue": max_fps,
                }
            )

        content = []
        for i, eval_result in enumerate(self.eval_results, 1):
            name = f"[{i}] {eval_result.name}"
            row = [name]
            for batch_size in batch_sizes:
                if batch_size in data[i]:
                    row.append(data[i][batch_size])
                else:
                    row.append("―")
            content.append(
                {
                    "row": row,
                    "id": name,
                    "items": row,
                }
            )
        data = {
            "columns": columns,
            "columnsOptions": columns_options,
            "content": content,
        }
        return TableWidget(
            name="fps_table",
            data=data,
            show_header_controls=False,
            fix_columns=1,
        )

    @property
    def inference_time_table(self) -> TableWidget:
        data = {}
        batch_sizes = set()
        for i, eval_result in enumerate(self.eval_results, 1):
            data[i] = {}
            if eval_result.speedtest_info is None:
                continue
            speedtests = eval_result.speedtest_info["speedtest"]
            for test in speedtests:
                batch_size = test["batch_size"]
                ms = round(test["benchmark"]["total"], 2)
                batch_sizes.add(batch_size)
                data[i][batch_size] = ms

        batch_sizes = sorted(batch_sizes)
        columns = ["Model"]
        columns_options = [{"disableSort": True}]
        for batch_size in batch_sizes:
            columns.extend([f"Batch size {batch_size}"])
            columns_options.extend(
                [
                    {"subtitle": "ms", "tooltip": "Milliseconds for batch images", "postfix": "ms"},
                ]
            )

        content = []
        for i, eval_result in enumerate(self.eval_results, 1):
            name = f"[{i}] {eval_result.name}"
            row = [name]
            for batch_size in batch_sizes:
                if batch_size in data[i]:
                    row.append(data[i][batch_size])
                else:
                    row.append("―")
            content.append(
                {
                    "row": row,
                    "id": name,
                    "items": row,
                }
            )

        data = {
            "columns": columns,
            "columnsOptions": columns_options,
            "content": content,
        }
        return TableWidget(
            name="inference_time_md",
            data=data,
            show_header_controls=False,
            fix_columns=1,
        )

    @property
    def batch_inference_md(self):
        return MarkdownWidget(
            name="batch_inference",
            title="Batch Inference",
            text=self.vis_texts.markdown_batch_inference,
        )

    @property
    def chart(self) -> ChartWidget:
        return ChartWidget(name="speed_charts", figure=self.get_figure())

    def get_figure(self):  #  -> Optional[go.Figure]
        import plotly.graph_objects as go  # pylint: disable=import-error
        from plotly.subplots import make_subplots  # pylint: disable=import-error

        fig = make_subplots(cols=2)

        for eval_result in self.eval_results:
            if eval_result.speedtest_info is None:
                continue
            temp_res = {}
            for test in eval_result.speedtest_info["speedtest"]:
                batch_size = test["batch_size"]

                std = test["benchmark_std"]["total"]
                ms = test["benchmark"]["total"]
                fps = round(1000 / test["benchmark"]["total"] * batch_size)

                ms_line = temp_res.setdefault("ms", {})
                fps_line = temp_res.setdefault("fps", {})
                ms_std_line = temp_res.setdefault("ms_std", {})

                ms_line[batch_size] = ms
                fps_line[batch_size] = fps
                ms_std_line[batch_size] = round(std, 2)

            error_color = "rgba(" + ",".join(map(str, hex2rgb(eval_result.color))) + ", 0.5)"
            fig.add_trace(
                go.Scatter(
                    x=list(temp_res["ms"].keys()),
                    y=list(temp_res["ms"].values()),
                    name="Infrence time (ms)",
                    line=dict(color=eval_result.color),
                    customdata=list(temp_res["ms_std"].values()),
                    error_y=dict(
                        type="data",
                        array=list(temp_res["ms_std"].values()),
                        visible=True,
                        color=error_color,
                    ),
                    hovertemplate="Batch Size: %{x}<br>Time: %{y:.2f} ms<br> Standard deviation: %{customdata:.2f} ms<extra></extra>",
                ),
                col=1,
                row=1,
            )
            fig.add_trace(
                go.Scatter(
                    x=list(temp_res["fps"].keys()),
                    y=list(temp_res["fps"].values()),
                    name="FPS",
                    line=dict(color=eval_result.color),
                    hovertemplate="Batch Size: %{x}<br>FPS: %{y:.2f}<extra></extra>",  # <br> Standard deviation: %{customdata:.2f}<extra></extra>",
                ),
                col=2,
                row=1,
            )

        fig.update_xaxes(title_text="Batch size", col=1, dtick=1)
        fig.update_xaxes(title_text="Batch size", col=2, dtick=1)

        fig.update_yaxes(title_text="Time (ms)", col=1)
        fig.update_yaxes(title_text="FPS", col=2)
        fig.update_layout(height=400)

        return fig



from typing import List

from supervisely import (
    AnyGeometry,
    Bitmap,
    GraphNodes,
    ObjClass,
    Point,
    Polygon,
    Polyline,
    ProjectMeta,
    Rectangle,
    TagMeta,
    TagValueType,
    logger,
)
from supervisely.geometry.graph import KeypointsTemplate
from supervisely.io.json import load_json_file
from supervisely.annotation.label import LabelJsonFields
from supervisely.annotation.tag import TagJsonFields
from supervisely.video_annotation.constants import FIGURES, INDEX, KEY, IMG_SIZE, OBJECT_KEY, OBJECTS, FRAMES_COUNT, TAGS, FRAMES

SLY_ANN_KEYS = [IMG_SIZE, FRAMES_COUNT, FRAMES, OBJECTS, TAGS]
SLY_VIDEO_OBJECT_KEYS = [LabelJsonFields.OBJ_CLASS_NAME, LabelJsonFields.TAGS, KEY]
SLY_TAG_KEYS = [TagJsonFields.TAG_NAME, TagJsonFields.VALUE]
SLY_FRAME_KEYS = [FIGURES, INDEX]
SLY_FIGURE_KEYS = [KEY, OBJECT_KEY, "geometryType"]  #, LabelJsonFields.GEOMETRY_TYPE] TODO: add geometry type


def get_meta_from_annotation(ann_path: str, meta: ProjectMeta) -> ProjectMeta:
    ann_json = load_json_file(ann_path)
    if "annotation" in ann_json:
        ann_json = ann_json["annotation"]
    if not all(key in ann_json for key in SLY_ANN_KEYS):
        logger.warn(
            f"VideoAnnotation file {ann_path} is not in Supervisely format. "
            "Check the annotation format documentation at: "
            "https://docs.supervisely.com/customization-and-integration/00_ann_format_navi/06_supervisely_format_videos"
        )
        return meta

    object_key_to_name = {}
    for object in ann_json[OBJECTS]:
        if not all(key in object for key in SLY_VIDEO_OBJECT_KEYS):
            logger.warn(
                f"Object in annotation file is not in Supervisely format: {object}. "
                "Read more about the Supervisely JSON format of objects in the documentation at: "
                "https://docs.supervisely.com/customization-and-integration/00_ann_format_navi/06_supervisely_format_videos"
            )
            continue
        meta = create_tags_from_annotation(meta, object[TAGS])
        object_key_to_name[object[KEY]] = object[LabelJsonFields.OBJ_CLASS_NAME]
    for frame in ann_json[FRAMES]:
        if not all(key in frame for key in SLY_FRAME_KEYS):
            logger.warn(
                f"Frame in annotation file is not in Supervisely format: {frame}."
                "Read more about the Supervisely JSON format of frames in the documentation at: "
                "https://docs.supervisely.com/customization-and-integration/00_ann_format_navi/06_supervisely_format_videos"
            )
            continue
        meta = create_classes_from_annotation(frame, meta, object_key_to_name)
    meta = create_tags_from_annotation(meta, ann_json[TAGS])
    return meta


def create_tags_from_annotation(meta: ProjectMeta, tags: List[dict]) -> ProjectMeta:
    for tag in tags:
        tag_name = tag[TagJsonFields.TAG_NAME]
        tag_value = tag.get(TagJsonFields.VALUE)
        if tag_value is None:
            tag_meta = TagMeta(tag_name, TagValueType.NONE)
        elif isinstance(tag_value, int) or isinstance(tag_value, float):
            tag_meta = TagMeta(tag_name, TagValueType.ANY_NUMBER)
        else:
            tag_meta = TagMeta(tag_name, TagValueType.ANY_STRING)

        # check existing tag_meta in meta
        existing_tag = meta.get_tag_meta(tag_name)
        if existing_tag is None:
            meta = meta.add_tag_meta(tag_meta)
    return meta


def create_classes_from_annotation(
    frame: dict, meta: ProjectMeta, object_key_to_name: dict
) -> ProjectMeta:
    for fig in frame[FIGURES]:
        if not all(key in fig for key in SLY_FIGURE_KEYS):
            logger.warn(
                f"Figure in annotation file is not in Supervisely format: {fig}. "
                "Read more about the Supervisely JSON format of figures in the documentation at: "
                "https://docs.supervisely.com/customization-and-integration/00_ann_format_navi/06_supervisely_format_videos"
            )
            continue
        obj_key = fig.get(OBJECT_KEY)
        if obj_key is None:
            continue
        class_name = object_key_to_name[obj_key]
        geometry_type = fig["geometryType"]
        obj_class = None
        if geometry_type == Bitmap.geometry_name():
            obj_class = ObjClass(name=class_name, geometry_type=Bitmap)
        elif geometry_type == Rectangle.geometry_name():
            obj_class = ObjClass(name=class_name, geometry_type=Rectangle)
        elif geometry_type == Point.geometry_name():
            obj_class = ObjClass(name=class_name, geometry_type=Point)
        elif geometry_type == Polygon.geometry_name():
            obj_class = ObjClass(name=class_name, geometry_type=Polygon)
        elif geometry_type == Polyline.geometry_name():
            obj_class = ObjClass(name=class_name, geometry_type=Polyline)
        elif geometry_type == GraphNodes.geometry_name():
            if not all(key in fig for key in ["geometry", "geometryType"]):
                continue
            geometry = fig["geometry"]
            if "nodes" not in geometry:
                continue
            template = KeypointsTemplate()
            for uuid, node in geometry["nodes"].items():
                if "loc" not in node or len(node["loc"]) != 2:
                    continue
                template.add_point(label=uuid, row=node["loc"][0], col=node["loc"][1])
            obj_class = ObjClass(name=class_name, geometry_type=GraphNodes, geometry_config=template)

        existing_class = meta.get_obj_class(class_name)
        if obj_class is None:
            logger.warn(f"Object class {class_name} is not in Supervisely format.")
            continue
        if existing_class is None:
            meta = meta.add_obj_class(obj_class)
        else:
            if existing_class.geometry_type != obj_class.geometry_type:
                obj_class = ObjClass(name=class_name, geometry_type=AnyGeometry)
                meta = meta.delete_obj_class(class_name)
                meta = meta.add_obj_class(obj_class)
    return meta

def rename_in_json(ann_json, renamed_classes=None, renamed_tags=None):
    if renamed_classes:
        for obj in ann_json[OBJECTS]:
            obj_cls_name = obj[LabelJsonFields.OBJ_CLASS_NAME]
            obj[LabelJsonFields.OBJ_CLASS_NAME] = renamed_classes.get(obj_cls_name, obj_cls_name)
            for tag in obj[TAGS]:
                tag_name = tag[TagJsonFields.TAG_NAME]
                tag[TagJsonFields.TAG_NAME] = renamed_tags.get(tag_name, tag_name)
    if renamed_tags:
        for tag in ann_json[TAGS]:
            tag_name = tag[TagJsonFields.TAG_NAME]
            tag[TagJsonFields.TAG_NAME] = renamed_tags.get(tag_name, tag_name)
    return ann_json



from supervisely.nn.inference import SessionJSON

WORKSPACE_NAME = "Model Benchmark: predictions and differences"
WORKSPACE_DESCRIPTION = "Technical workspace for model benchmarking. Contains predictions and differences between ground truth and predictions."

def try_set_conf_auto(session: SessionJSON, conf: float):
    conf_names = ["conf", "confidence", "confidence_threshold", "confidence_thresh"]
    default = session.get_default_inference_settings()
    for name in conf_names:
        if name in default:
            session.inference_settings[name] = conf
            return True
    return False



from __future__ import annotations

from typing import TYPE_CHECKING

from supervisely.nn.benchmark.visualization.vis_metric_base import MetricVis
from supervisely.nn.benchmark.visualization.vis_widgets import Schema, Widget

if TYPE_CHECKING:
    from supervisely.nn.benchmark.visualization.visualizer import Visualizer


class SpeedtestIntro(MetricVis):

    def __init__(self, loader: Visualizer) -> None:
        super().__init__(loader)
        speedtest = self._loader.speedtest["speedtest"][0]
        self.schema = Schema(
            self._loader.inference_speed_text,
            markdown_speedtest_intro=Widget.Markdown(
                title="Inference Speed",
                is_header=True,
                formats=[
                    speedtest["device"],
                    self._loader.hardware,
                    speedtest["runtime"]
                ],
            ),
        )



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

if TYPE_CHECKING:
    from supervisely.nn.benchmark.visualization.visualizer import Visualizer

from jinja2 import Template

from supervisely._utils import camel_to_snake
from supervisely.nn.benchmark.cv_tasks import CVTask
from supervisely.nn.benchmark.visualization.vis_templates import (
    template_chart_str,
    template_gallery_str,
    template_markdown_str,
    template_notification_str,
    template_radiogroup_str,
    template_table_str,
)
from supervisely.nn.benchmark.visualization.vis_widgets import Schema, Widget
from supervisely.project.project_meta import ProjectMeta


class MetricVis:

    def __init__(self, loader: Visualizer) -> None:

        self.cv_tasks: List[CVTask] = CVTask.values()
        self.clickable: bool = False
        self.has_diffs_view: bool = False
        self.switchable: bool = False
        self.schema: Schema = None
        self.empty = False
        self._is_overview = False

        self._loader = loader
        self._template_markdown = Template(template_markdown_str)
        self._template_chart = Template(template_chart_str)
        self._template_radiogroup = Template(template_radiogroup_str)
        self._template_gallery = Template(template_gallery_str)
        self._template_table = Template(template_table_str)
        self._template_notification = Template(template_notification_str)
        self._keypair_sep = "-"

    @property
    def radiogroup_id(self) -> Optional[str]:
        if self.switchable:
            return f"radiogroup_" + self.name

    @property
    def template_sidebar_str(self) -> str:
        res = ""
        for widget in self.schema:  # pylint: disable=not-an-iterable
            if isinstance(widget, Widget.Markdown):
                if widget.title is not None and widget.is_header:
                    res += f"""\n          <div>\n            <el-button type="text" @click="data.scrollIntoView='{widget.id}'" """
                    res += (
                        """:style="{fontWeight: data.scrollIntoView==='"""
                        + widget.id
                        + """' ? 'bold' : 'normal'}"""
                    )
                    res += f""" ">{widget.title}</el-button>\n          </div>"""
        return res

    @property
    def template_main_str(self) -> str:
        res = ""

        _is_before_chart = True

        def _add_radio_buttons():
            res = ""
            for widget in self.schema:  # pylint: disable=not-an-iterable
                if isinstance(widget, Widget.Chart):
                    basename = f"{widget.name}_{self.name}"
                    res += "\n            {{ " + f"el_radio_{basename}_html" + " }}"
            return res

        is_radiobuttons_added = False

        for widget in self.schema:  # pylint: disable=not-an-iterable
            if isinstance(widget, Widget.Chart):
                _is_before_chart = False

            if (
                isinstance(widget, (Widget.Markdown, Widget.Notification, Widget.Collapse))
                and _is_before_chart
            ):
                res += "\n            {{ " + f"{widget.name}_html" + " }}"
                continue

            if isinstance(widget, (Widget.Chart, Widget.Gallery, Widget.Table)):
                basename = f"{widget.name}_{self.name}"
                if self.switchable and not is_radiobuttons_added:
                    res += _add_radio_buttons()
                    is_radiobuttons_added = True
                res += "\n            {{ " + f"{basename}_html" + " }}"
                if self.clickable:
                    res += "\n            {{ " + f"{basename}_clickdata_html" + " }}"
                continue

            if (
                isinstance(widget, (Widget.Markdown, Widget.Notification, Widget.Collapse))
                and not _is_before_chart
            ):
                res += "\n            {{ " + f"{widget.name}_html" + " }}"
                continue

        return res

    def get_html_snippets(self) -> dict:
        res = {}
        for widget in self.schema:  # pylint: disable=not-an-iterable
            if isinstance(widget, Widget.Markdown):
                res[f"{widget.name}_html"] = self._template_markdown.render(
                    {
                        "widget_id": widget.id,
                        "data_source": f"/data/{widget.name}.md",
                        "command": "command",
                        "data": "data",
                        "is_overview": widget.title == "Overview",
                    }
                )

            if isinstance(widget, Widget.Collapse):
                subres = {}
                for subwidget in widget.schema:
                    if isinstance(subwidget, Widget.Markdown):
                        subres[f"{subwidget.name}_html"] = self._template_markdown.render(
                            {
                                "widget_id": subwidget.id,
                                "data_source": f"/data/{subwidget.name}.md",
                                "command": "command",
                                "data": "data",
                            }
                        )
                res[f"{widget.name}_html"] = widget.template_schema.render(**subres)
                continue

            if isinstance(widget, Widget.Notification):
                res[f"{widget.name}_html"] = self._template_notification.render(
                    {
                        "widget_id": widget.id,
                        "data": "data",
                        "title": widget.title.format(*widget.formats_title),
                        "description": widget.description.format(*widget.formats_desc),
                    }
                )

            if isinstance(widget, Widget.Chart):
                basename = f"{widget.name}_{self.name}"
                if self.switchable:
                    res[f"el_radio_{basename}_html"] = self._template_radiogroup.render(
                        {
                            "radio_group": self.radiogroup_id,
                            "switch_key": widget.switch_key,
                        }
                    )
                chart_click_path = f"/data/{basename}_click_data.json" if self.clickable else None
                chart_modal_data_source = f"/data/modal_general.json" if self.clickable else None
                res[f"{basename}_html"] = self._template_chart.render(
                    {
                        "widget_id": widget.id,
                        "init_data_source": f"/data/{basename}.json",
                        "chart_click_data_source": chart_click_path,
                        "command": "command",
                        "data": "data",
                        "cls_name": self.name,
                        "key_separator": self._keypair_sep,
                        "switchable": self.switchable,
                        "radio_group": self.radiogroup_id,
                        "switch_key": widget.switch_key,
                        "chart_modal_data_source": chart_modal_data_source,
                    }
                )
            if isinstance(widget, Widget.Gallery):
                basename = f"{widget.name}_{self.name}"
                if widget.is_table_gallery:
                    for w in self.schema:  # pylint: disable=not-an-iterable
                        if isinstance(w, Widget.Table):
                            w.gallery_id = widget.id

                gallery_click_data_source = (
                    f"/data/{basename}_click_data.json" if self.clickable else None
                )
                gallery_modal_data_source = (
                    f"/data/{basename}_modal_data.json" if self.clickable else None
                )
                gallery_diff_data_source = (
                    f"/data/{basename}_diff_data.json" if self.has_diffs_view else None
                )
                res[f"{basename}_html"] = self._template_gallery.render(
                    {
                        "widget_id": widget.id,
                        "init_data_source": f"/data/{basename}.json",
                        "command": "command",
                        "data": "data",
                        "is_table_gallery": widget.is_table_gallery,
                        "gallery_click_data_source": gallery_click_data_source,
                        "gallery_diff_data_source": gallery_diff_data_source,
                        "gallery_modal_data_source": gallery_modal_data_source,
                    }
                )

            if isinstance(widget, Widget.Table):
                basename = f"{widget.name}_{self.name}"
                res[f"{basename}_html"] = self._template_table.render(
                    {
                        "widget_id": widget.id,
                        "init_data_source": f"/data/{basename}.json",
                        "command": "command",
                        "data": "data",
                        "table_click_data": f"/data/{widget.name}_{self.name}_click_data.json",
                        "table_gallery_id": f"modal_general",
                        "clickable": self.clickable,
                        "mainColumn": widget.main_column,
                        "fixColumns": widget.fixed_columns,
                        "showHeaderControls": widget.show_header_controls,
                        "width": f"width:{widget.width}" if widget.width else "",
                    }
                )

        return res

    @property
    def name(self) -> str:
        return camel_to_snake(self.__class__.__name__)

    @property
    def f1_optimal_conf(self) -> Optional[float]:
        return self._loader.f1_optimal_conf

    def get_figure(self, widget: Widget.Chart):  # -> Optional[go.Figure]:
        pass

    def get_click_data(self, widget: Widget.Chart) -> Optional[dict]:
        if not self.clickable:
            return

        res = {}

        res["layoutTemplate"] = [None, None, None]
        res["clickData"] = {}
        for key, v in self._loader.click_data.objects_by_class.items():
            res["clickData"][key] = {}
            res["clickData"][key]["imagesIds"] = []

            # tmp = defaultdict(list)
            img_ids = set()
            obj_ids = set()

            res["clickData"][key][
                "title"
            ] = f"{key} class: {len(v)} object{'s' if len(v) > 1 else ''}"

            for x in v:
                img_ids.add(x["dt_img_id"])
                obj_ids.add(x["dt_obj_id"])

            res["clickData"][key]["imagesIds"] = list(img_ids)
            res["clickData"][key]["filters"] = [
                {"type": "tag", "tagId": "confidence", "value": [self.f1_optimal_conf, 1]},
                {"type": "tag", "tagId": "outcome", "value": "TP"},
                {"type": "specific_objects", "tagId": None, "value": list(obj_ids)},
            ]

        return res

    def get_modal_data(self, widget: Widget.Chart) -> Optional[dict]:
        res = {}
        api = self._loader._api
        gt_project_id = self._loader.gt_project_info.id
        dt_project_id = self._loader.dt_project_info.id
        diff_project_id = self._loader.diff_project_info.id
        gt_dataset = api.dataset.get_list(gt_project_id)[0]
        dt_dataset = api.dataset.get_list(dt_project_id)[0]
        diff_dataset = api.dataset.get_list(diff_project_id)[0]
        gt_image_infos = api.image.get_list(dataset_id=gt_dataset.id)[:3]
        pred_image_infos = api.image.get_list(dataset_id=dt_dataset.id)[:3]
        diff_image_infos = api.image.get_list(dataset_id=diff_dataset.id)[:3]
        project_metas = [
            ProjectMeta.from_json(data=api.project.get_meta(id=x))
            for x in [gt_project_id, dt_project_id, diff_project_id]
        ]
        for gt_image, pred_image, diff_image in zip(
            gt_image_infos, pred_image_infos, diff_image_infos
        ):
            image_infos = [gt_image, pred_image, diff_image]
            ann_infos = [api.annotation.download(x.id) for x in image_infos]

            for idx, (image_info, ann_info, project_meta) in enumerate(
                zip(image_infos, ann_infos, project_metas)
            ):
                image_name = image_info.name
                image_url = image_info.preview_url
                is_ignore = True if idx in [0, 1] else False
                widget.gallery.append(
                    title=image_name,
                    image_url=image_url,
                    annotation_info=ann_info,
                    column_index=idx,
                    project_meta=project_meta,
                    ignore_tags_filtering=is_ignore,
                )

        res.update(widget.gallery.get_json_state())
        res.update(widget.gallery.get_json_data()["content"])
        res["layoutData"] = res.pop("annotations")
        res["projectMeta"] = project_metas[0].to_json()

        res.pop("layout")
        res.pop("layoutData")

        return res

    def get_table(self, widget: Widget.Table) -> Optional[dict]:
        pass

    def get_gallery(self, widget: Widget.Gallery) -> Optional[dict]:
        pass

    def get_gallery_click_data(self, widget: Widget.Gallery) -> Optional[dict]:
        pass

    def get_diff_gallery_data(self, widget: Widget.Gallery) -> Optional[dict]:
        pass

    def get_md_content(self, widget: Widget.Markdown):
        if hasattr(self._loader.vis_texts, widget.name):
            return getattr(self._loader.vis_texts, widget.name).format(*widget.formats)
        elif hasattr(self._loader.inference_speed_text, widget.name):
            return getattr(self._loader.inference_speed_text, widget.name).format(*widget.formats)
        else:
            raise AttributeError(f"Not found texts template for {widget.name}")

    def initialize_formats(self, loader: Visualizer, widget: Widget):
        pass



from supervisely.nn.benchmark.base_benchmark import BaseBenchmark
from supervisely.nn.benchmark.cv_tasks import CVTask
from supervisely.nn.benchmark.evaluation.object_detection_evaluator import (
    ObjectDetectionEvaluator,
)
from supervisely.nn.benchmark.utils import try_set_conf_auto

CONF_THRES = 0.05


class ObjectDetectionBenchmark(BaseBenchmark):

    @property
    def cv_task(self) -> str:
        return CVTask.OBJECT_DETECTION

    def _get_evaluator_class(self) -> type:
        return ObjectDetectionEvaluator

    def _run_inference(
        self,
        output_project_id=None,
        batch_size: int = 8,
        cache_project_on_agent: bool = False,
    ):
        assert try_set_conf_auto(
            self.session, CONF_THRES
        ), f"Unable to set the confidence threshold to {CONF_THRES} for evalation."
        return super()._run_inference(output_project_id, batch_size, cache_project_on_agent)



# coding: utf-8
from copy import deepcopy

from supervisely import logger
from supervisely.function_wrapper import function_wrapper
from supervisely.io.json import load_json_file
from supervisely.nn.legacy.config import update_recursively
from supervisely.nn.legacy.hosted.constants import CONFIG, MODEL
from supervisely.task.paths import TaskPaths

from supervisely.worker_api.rpc_servicer import AgentRPCServicer
from supervisely.worker_api.agent_rpc import SimpleCache


class ModelDeploy:

    config = {
        'cache_limit': 500,
        'connection': {
            'server_address': None,
            'token': None,
            'task_id': None,
        },
        MODEL: {
        }
    }

    def __init__(self, model_applier_cls, rpc_servicer_cls=AgentRPCServicer):
        self.model_applier_cls = model_applier_cls
        self.rpc_servicer_cls = rpc_servicer_cls
        self.load_config()
        self._create_serv_instance()

    def load_config(self):
        self.config = deepcopy(ModelDeploy.config)
        new_config = load_json_file(TaskPaths.TASK_CONFIG_PATH)
        logger.info('Input config', extra={CONFIG: new_config})
        update_recursively(self.config, new_config)
        logger.info('Full config', extra={CONFIG: self.config})

    def _create_applier(self):
        model_applier = function_wrapper(self.model_applier_cls)
        return model_applier

    def _create_serv_instance(self):
        model_applier = self._create_applier()
        image_cache = SimpleCache(self.config['cache_limit'])
        self.serv_instance = self.rpc_servicer_cls(logger=logger,
                                                   model_applier=model_applier,
                                                   conn_config=self.config['connection'],
                                                   cache=image_cache)

    def run(self):
        self.serv_instance.run_inf_loop()



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from supervisely.nn.benchmark.visualization.vis_metric_base import MetricVis
from supervisely.nn.benchmark.visualization.vis_widgets import Schema, Widget

if TYPE_CHECKING:
    from supervisely.nn.benchmark.visualization.visualizer import Visualizer


class PRCurve(MetricVis):

    def __init__(self, loader: Visualizer) -> None:
        super().__init__(loader)
        self.schema = Schema(
            self._loader.vis_texts,
            markdown_pr_curve=Widget.Markdown(
                title="Precision-Recall Curve",
                is_header=True,
                formats=[self._loader.vis_texts.definitions.f1_score],
            ),
            collapse_pr=Widget.Collapse(
                schema=Schema(
                    self._loader.vis_texts,
                    markdown_trade_offs=Widget.Markdown(
                        title="About Trade-offs between precision and recall"
                    ),
                    markdown_what_is_pr_curve=Widget.Markdown(
                        title="How the PR curve is built?",
                        formats=[
                            self._loader.vis_texts.definitions.confidence_score,
                            self._loader.vis_texts.definitions.true_positives,
                            self._loader.vis_texts.definitions.false_positives,
                        ],
                    ),
                )
            ),
            notification_ap=Widget.Notification(
                formats_title=[loader.base_metrics()["mAP"].round(2)]
            ),
            chart=Widget.Chart(),
        )

    def get_figure(self, widget: Widget.Chart):  # -> Optional[go.Figure]:
        import plotly.express as px  # pylint: disable=import-error
        import plotly.graph_objects as go  # pylint: disable=import-error

        # Precision-Recall curve
        pr_curve = self._loader.mp.pr_curve().copy()
        pr_curve[pr_curve == -1] = np.nan  # -1 is a placeholder for no GT
        pr_curve = np.nanmean(pr_curve, axis=-1)
        fig = px.line(
            x=self._loader.mp.recThrs,
            y=pr_curve,
            # title="Precision-Recall Curve",
            labels={"x": "Recall", "y": "Precision"},
            width=600,
            height=500,
        )
        fig.data[0].name = "Model"
        fig.data[0].showlegend = True
        fig.update_traces(fill="tozeroy", line=dict(color="#1f77b4"))
        fig.add_trace(
            go.Scatter(
                x=self._loader.mp.recThrs,
                y=[1] * len(self._loader.mp.recThrs),
                name="Perfect",
                line=dict(color="orange", dash="dash"),
                showlegend=True,
            )
        )
        fig.add_annotation(
            text=f"mAP = {self._loader.mp.base_metrics()['mAP']:.2f}",
            xref="paper",
            yref="paper",
            x=0.98,
            y=0.92,
            showarrow=False,
            bgcolor="white",
        )
        fig.update_traces(hovertemplate="Recall: %{x:.2f}<br>Precision: %{y:.2f}<extra></extra>")
        fig.update_layout(
            dragmode=False,
            modebar=dict(
                remove=[
                    "zoom2d",
                    "pan2d",
                    "select2d",
                    "lasso2d",
                    "zoomIn2d",
                    "zoomOut2d",
                    "autoScale2d",
                    "resetScale2d",
                ]
            ),
        )
        # fig.show()
        return fig



import traceback

from rich.console import Console
import tqdm
import supervisely as sly


def download_run(id: int, dest_dir: str) -> bool:
    console = Console()

    api = sly._handle_creds_error_to_console(sly.Api.from_env, console.print)
    if api is False:
        return False

    console.print(
        f"\nDownloading data from project with ID={id} to directory: '{dest_dir}' ...\n",
        style="bold",
    )

    project_info = api.project.get_info_by_id(id)
    if project_info is None:
        console.print(f"\nError: Project '{project_info}' doesn't exists\n", style="bold red")
        return False

    n_count = project_info.items_count
    try:
        with tqdm.tqdm(total=n_count) as pbar:
            sly.download(api, id, dest_dir, progress_cb=pbar)

        console.print(
            f"\nProject '{project_info.name}' has been downloaded sucessfully!\n",
            style="bold green",
        )
        return True
    except:
        console.print(
            f"\nProject '{project_info.name}' has been failed to download\n", style="bold red"
        )
        traceback.print_exc()
        return False



from __future__ import annotations

from typing import TYPE_CHECKING

from supervisely.nn.benchmark.visualization.vis_metric_base import MetricVis
from supervisely.nn.benchmark.visualization.vis_widgets import Schema, Widget

if TYPE_CHECKING:
    from supervisely.nn.benchmark.visualization.visualizer import Visualizer


class RecallVsPrecision(MetricVis):

    def __init__(self, loader: Visualizer) -> None:
        super().__init__(loader)
        self.clickable = True
        self.schema = Schema(
            self._loader.vis_texts,
            markdown_PR=Widget.Markdown(
                title="Recall vs Precision",
                is_header=True,
                formats=[self._loader.vis_texts.definitions.f1_score],
            ),
            chart=Widget.Chart(),
        )

    def get_figure(self, widget: Widget.Chart):  #  -> Optional[go.Figure]
        import plotly.graph_objects as go  # pylint: disable=import-error

        blue_color = "#1f77b4"
        orange_color = "#ff7f0e"
        sorted_by_f1 = self._loader.mp.per_class_metrics().sort_values(by="f1")
        fig = go.Figure()
        fig.add_trace(
            go.Bar(
                y=sorted_by_f1["precision"],
                x=sorted_by_f1["category"],
                name="Precision",
                marker=dict(color=blue_color),
            )
        )
        fig.add_trace(
            go.Bar(
                y=sorted_by_f1["recall"],
                x=sorted_by_f1["category"],
                name="Recall",
                marker=dict(color=orange_color),
            )
        )
        fig.update_layout(
            barmode="group",
            # title="Per-class Precision and Recall (Sorted by F1)",
        )
        fig.update_xaxes(title_text="Class")
        fig.update_yaxes(title_text="Value", range=[0, 1])
        # fig.show()
        return fig



from __future__ import annotations

from pathlib import Path
from typing import Dict, List, Tuple, Union

import numpy as np
import requests

from supervisely import Annotation, Label
from supervisely.nn.tracker.tracker import BaseDetection as Detection
from supervisely.nn.tracker.tracker import BaseTrack, BaseTracker
from supervisely.sly_logger import logger

from . import matching
from .bot_sort import (
    BoTSORT,
    STrack,
    TrackState,
    joint_stracks,
    remove_duplicate_stracks,
    sub_stracks,
)


class Track(BaseTrack, STrack):
    def __init__(self, tlwh, confidence: float, feature=None, sly_label: Label = None):
        self.track_id = None
        STrack.__init__(self, tlwh, confidence, sly_label.obj_class.name, feature)
        self.sly_label = sly_label

    def get_sly_label(self):
        return self.sly_label

    def update(self, new: Track, frame_id):
        STrack.update(self, new, frame_id)
        self.sly_label = new.get_sly_label()

    def re_activate(self, new: Track, frame_id, new_id=False):
        STrack.re_activate(self, new, frame_id, new_id)
        self.sly_label = new.get_sly_label()


class _BoTSORT(BoTSORT):
    def __init__(self, args):
        super().__init__(args)

        # for type hinting
        self.tracked_stracks = []  # type: list[Track]
        self.lost_stracks = []  # type: list[Track]
        self.removed_stracks = []  # type: list[Track]

        # for pylint
        self.frame_id = 0

    def _init_track(
        self, dets: np.ndarray, scores: np.ndarray, labels: List[Label], features: np.ndarray
    ):
        if len(dets) == 0:
            return []
        if self.args.with_reid:
            return [
                Track(Track.tlbr_to_tlwh(tlbr), s, f, l)
                for (tlbr, s, f, l) in zip(dets, scores, features, labels)
            ]
        return [
            Track(Track.tlbr_to_tlwh(tlbr), s, None, l)
            for (tlbr, s, l) in zip(dets, scores, labels)
        ]

    def get_dists(self, tracks: List[Track], detections: List[Track]):
        # Associate with high score detection boxes
        ious_dists = matching.iou_distance(tracks, detections)

        # if not self.args.mot20:
        ious_dists = matching.fuse_score(ious_dists, detections)

        # Remove detections with different shapes
        for track_i, track in enumerate(tracks):
            for det_i, det in enumerate(detections):
                if track.get_sly_label().geometry.name() != det.get_sly_label().geometry.name():
                    ious_dists[track_i, det_i] = 1.0

        if not self.args.with_reid:
            return ious_dists

        ious_dists_mask = ious_dists > self.args.proximity_thresh
        emb_dists = matching.embedding_distance(tracks, detections) / 2.0
        # raw_emb_dists = emb_dists.copy()
        emb_dists[emb_dists > self.args.appearance_thresh] = 1.0
        emb_dists[ious_dists_mask] = 1.0
        dists = np.minimum(ious_dists, emb_dists)

        # Popular ReID method (JDE / FairMOT)
        # raw_emb_dists = matching.embedding_distance(strack_pool, detections)
        # dists = matching.fuse_motion(self.kalman_filter, raw_emb_dists, strack_pool, detections)
        # emb_dists = dists

        # IoU making ReID
        # dists = matching.embedding_distance(strack_pool, detections)
        # dists[ious_dists_mask] = 1.0
        return dists

    def update(self, detections_: List[Detection], img) -> List[Track]:
        self.frame_id += 1
        activated_starcks = []
        refind_stracks = []
        lost_stracks = []
        removed_stracks = []

        if len(detections_):
            scores = np.array([det.confidence for det in detections_])
            bboxes = np.array([det.tlbr()[:4] for det in detections_])
            labels = np.array([det.sly_label for det in detections_])
            features = np.array([det.feature for det in detections_])

            # Remove bad detections
            lowest_inds = scores > self.args.track_low_thresh
            bboxes = bboxes[lowest_inds]
            scores = scores[lowest_inds]
            labels = labels[lowest_inds]
            features = features[lowest_inds]

            # Find high threshold detections
            remain_inds = scores > self.args.track_high_thresh
            dets = bboxes[remain_inds]
            scores_keep = scores[remain_inds]
            labels_keep = labels[remain_inds]
            features_keep = features[remain_inds]
        else:
            bboxes = []
            scores = []
            labels = []
            dets = []
            scores_keep = []
            labels_keep = []
            features_keep = []

        """Extract embeddings """
        if self.args.with_reid:
            features_keep = self.encoder.inference(img, dets)

        """Detections"""
        detections = self._init_track(dets, scores_keep, labels_keep, features_keep)

        """ Add newly detected tracklets to tracked_stracks"""
        unconfirmed = []
        tracked_stracks = []  # type: list[STrack]
        for track in self.tracked_stracks:
            if not track.is_activated:
                unconfirmed.append(track)
            else:
                tracked_stracks.append(track)

        """ Step 2: First association, with high score detection boxes"""
        strack_pool: List[Track] = joint_stracks(tracked_stracks, self.lost_stracks)

        # Predict the current location with KF
        STrack.multi_predict(strack_pool)

        # Fix camera motion
        warp = self.gmc.apply(img, dets)
        STrack.multi_gmc(strack_pool, warp)
        STrack.multi_gmc(unconfirmed, warp)

        dists = self.get_dists(strack_pool, detections)

        matches, u_track, u_detection = matching.linear_assignment(
            dists, thresh=self.args.match_thresh
        )

        for itracked, idet in matches:
            track: Track = strack_pool[itracked]
            det: Track = detections[idet]
            if track.state == TrackState.Tracked:
                track.update(detections[idet], self.frame_id)
                activated_starcks.append(track)
            else:
                track.re_activate(det, self.frame_id, new_id=False)
                refind_stracks.append(track)

        """ Step 3: Second association, with low score detection boxes"""
        if len(scores):
            inds_second = scores < self.args.track_high_thresh
            dets_second = bboxes[inds_second]
            scores_second = scores[inds_second]
            labels_second = labels[inds_second]
            features_second = features[inds_second]
        else:
            dets_second = []
            scores_second = []
            labels_second = []
            features_second = []

        # association the untrack to the low score detections
        detections_second = self._init_track(
            dets_second, scores_second, labels_second, features_second
        )

        r_tracked_stracks = [
            strack_pool[i] for i in u_track if strack_pool[i].state == TrackState.Tracked
        ]
        dists = matching.iou_distance(r_tracked_stracks, detections_second)
        matches, u_track, u_detection_second = matching.linear_assignment(dists, thresh=0.5)
        for itracked, idet in matches:
            track = r_tracked_stracks[itracked]
            det = detections_second[idet]
            if track.state == TrackState.Tracked:
                track.update(det, self.frame_id)
                activated_starcks.append(track)
            else:
                track.re_activate(det, self.frame_id, new_id=False)
                refind_stracks.append(track)

        for it in u_track:
            track = r_tracked_stracks[it]
            if not track.state == TrackState.Lost:
                track.mark_lost()
                lost_stracks.append(track)

        """Deal with unconfirmed tracks, usually tracks with only one beginning frame"""
        detections = [detections[i] for i in u_detection]
        dists = self.get_dists(unconfirmed, detections)
        matches, u_unconfirmed, u_detection = matching.linear_assignment(dists, thresh=0.7)
        for itracked, idet in matches:
            unconfirmed[itracked].update(detections[idet], self.frame_id)
            activated_starcks.append(unconfirmed[itracked])
        for it in u_unconfirmed:
            track = unconfirmed[it]
            track.mark_removed()
            removed_stracks.append(track)

        """ Step 4: Init new stracks"""
        for inew in u_detection:
            track = detections[inew]
            if track.score < self.new_track_thresh:
                continue

            track.activate(self.kalman_filter, self.frame_id)
            activated_starcks.append(track)

        """ Step 5: Update state"""
        for track in self.lost_stracks:
            if self.frame_id - track.end_frame > self.max_time_lost:
                track.mark_removed()
                removed_stracks.append(track)

        """ Merge """
        self.tracked_stracks = [t for t in self.tracked_stracks if t.state == TrackState.Tracked]
        self.tracked_stracks = joint_stracks(self.tracked_stracks, activated_starcks)
        self.tracked_stracks = joint_stracks(self.tracked_stracks, refind_stracks)
        self.lost_stracks = sub_stracks(self.lost_stracks, self.tracked_stracks)
        self.lost_stracks.extend(lost_stracks)
        self.lost_stracks = sub_stracks(self.lost_stracks, self.removed_stracks)
        self.removed_stracks.extend(removed_stracks)
        self.tracked_stracks, self.lost_stracks = remove_duplicate_stracks(
            self.tracked_stracks, self.lost_stracks
        )

        # output_stracks = [track for track in self.tracked_stracks if track.is_activated]
        output_stracks = [track for track in self.tracked_stracks]

        return output_stracks


class BoTTracker(BaseTracker):
    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        super().__init__(settings=settings)
        self.tracker = _BoTSORT(self.args)

        if self.args.with_reid:
            if not Path(self.args.fast_reid_weights).exists():
                logger.info("Downloading ReID weights...")

                with requests.get(self.args.fast_reid_weights_url, stream=True) as r:
                    r.raise_for_status()
                    with open(self.args.fast_reid_weights, "wb") as f:
                        for chunk in r.iter_content(chunk_size=8192):
                            f.write(chunk)

    def default_settings(self):
        return {
            "name": None,
            "ablation": None,
            # Tracking
            "track_high_thresh": 0.6,
            "track_low_thresh": 0.1,
            "new_track_thresh": 0.7,
            "track_buffer": 30,
            "match_thresh": 0.5,
            "min_box_area": 10,
            "fuse_score": False,
            # CMC
            "cmc_method": "sparseOptFlow",
            "gmc_config": None,
            # ReID
            "with_reid": False,
            "fast_reid_config": f"{Path(__file__).parent}/fast_reid/configs/MOT17/sbs_S50.yml",
            "fast_reid_weights": f"pretrained/yolo7x.pt",
            "fast_reid_weights_url": r"https://github.com/WongKinYiu/yolov7/releases/download/v0.1/yolov7.pt",
            "proximity_thresh": 0.5,
            "appearance_thresh": 0.25,
        }

    def track(
        self,
        source: Union[List[np.ndarray], List[str], str],
        frame_to_annotation: Dict[int, Annotation],
        frame_shape: Tuple[int, int],
        pbar_cb=None,
    ) -> Annotation:
        """
        Track objects in the video using BoTSort algorithm.

        :param source: List of images, paths to images or path to the video file.
        :type source: List[np.ndarray] | List[str] | str
        :param frame_to_annotation: Dictionary with frame index as key and Annotation as value.
        :type frame_to_annotation: Dict[int, Annotation]
        :param frame_shape: Size of the frame (height, width).
        :type frame_shape: Tuple[int, int]
        :param pbar_cb: Callback to update progress bar.
        :type pbar_cb: Callable, optional

        :return: Video annotation with tracked objects.
        :rtype: VideoAnnotation

        :raises ValueError: If number of images and annotations are not the same.

        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from supervisely.nn.tracker import BoTTracker

            api = sly.Api()

            project_id = 12345
            video_id = 12345678
            video_path = "video.mp4"

            # Download video and get video info
            video_info = api.video.get_info_by_id(video_id)
            frame_shape = (video_info.frame_height, video_info.frame_width)
            api.video.download_path(id=video_id, path=video_path)

            # Run inference app to get detections
            task_id = 12345 # detection app task id
            session = sly.nn.inference.Session(api, task_id)
            annotations = session.inference_video_id(video_id, 0, video_info.frames_count)
            frame_to_annotation = {i: ann for i, ann in enumerate(annotations)}

            # Run tracker
            tracker = BoTTracker()
            video_ann = tracker.track(video_path, frame_to_annotation, frame_shape)

            # Upload result
            model_meta = session.get_model_meta()
            project_meta = sly.ProjectMeta.from_json(api.project.get_meta(project_id))
            project_meta = project_meta.merge(model_meta)
            api.project.update_meta(project_id, project_meta)
            api.video.annotation.append(video_id, video_ann)

        """
        if not isinstance(source, str):
            if len(source) != len(frame_to_annotation):
                raise ValueError("Number of images and annotations should be the same")

        tracks_data = {}
        logger.info("Starting BoTSort tracking...")
        for frame_index, img in enumerate(self.frames_generator(source)):
            self.update(img, frame_to_annotation[frame_index], frame_index, tracks_data=tracks_data)

            if pbar_cb is not None:
                pbar_cb()

        return self.get_annotation(
            tracks_data=tracks_data,
            frame_shape=frame_shape,
            frames_count=len(frame_to_annotation),
        )

    def update(
        self, img, annotation: Annotation, frame_index, tracks_data: Dict[int, List[Dict]] = None
    ):
        pred, sly_labels = self.convert_annotation(annotation)

        detections = [Detection(p[:4], p[4], None, label) for p, label in zip(pred, sly_labels)]

        self.tracker.update(detections, img)

        if tracks_data is None:
            tracks_data = {}
        self.update_track_data(
            tracks_data=tracks_data,
            tracks=self.tracker.tracked_stracks,
            frame_index=frame_index,
        )

        return tracks_data



from __future__ import annotations

import os
from typing import Dict, List, Optional, Tuple, Union

from tqdm import tqdm

from supervisely._utils import is_production
from supervisely.annotation.annotation import Annotation
from supervisely.annotation.tag_meta import TagValueType
from supervisely.api.api import Api
from supervisely.io.fs import get_file_ext, get_file_name_with_ext
from supervisely.project.project_meta import ProjectMeta
from supervisely.project.project_settings import LabelingInterface
from supervisely.sly_logger import logger
from supervisely.task.progress import Progress


class AvailableImageConverters:
    SLY = "supervisely"
    COCO = "coco"
    FAST_COCO = "coco (fast)"
    YOLO = "yolo"
    PASCAL_VOC = "pascal_voc"
    CSV = "csv"
    MULTISPECTRAL = "multispectral"
    MASKS = "images_with_masks"
    MEDICAL2D = "medical_2d"
    MULTI_VIEW = "multi_view"
    PDF = "pdf"
    CITYSCAPES = "cityscapes"
    LABEL_ME = "label_me"
    LABEL_STUDIO = "label_studio"
    HIGH_COLOR_DEPTH = "high_color_depth"


class AvailableVideoConverters:
    SLY = "supervisely"
    MOT = "mot"
    DAVIS = "davis"


class AvailablePointcloudConverters:
    SLY = "supervisely"
    LAS = "las/laz"
    PLY = "ply"
    BAG = "rosbag"


class AvailablePointcloudEpisodesConverters:
    SLY = "supervisely"
    BAG = "rosbag"


class AvailableVolumeConverters:
    SLY = "supervisely"
    DICOM = "dicom"


class BaseConverter:
    unsupported_exts = [".gif", ".html", ".htm"]

    class BaseItem:
        def __init__(
            self,
            item_path: str,
            ann_data: Union[str, dict] = None,
            shape: Union[Tuple, List] = None,
            custom_data: dict = {},
        ):
            self._path: str = item_path
            self._name: str = None
            self._ann_data: Union[str, dict, list] = ann_data
            self._shape: Union[Tuple, List] = shape
            self._custom_data: dict = custom_data

        @property
        def name(self) -> str:
            if self._name is not None:
                return self._name
            return get_file_name_with_ext(self._path)

        @name.setter
        def name(self, name: str) -> None:
            self._name = name

        @property
        def path(self) -> str:
            return self._path

        @path.setter
        def path(self, path: str) -> None:
            self._path = path

        @property
        def ann_data(self) -> Union[str, dict]:
            return self._ann_data

        @ann_data.setter
        def ann_data(self, ann_data: Union[str, dict, list]) -> None:
            self._ann_data = ann_data

        @property
        def shape(self) -> Union[Tuple, List]:
            return self._shape

        @property
        def custom_data(self) -> dict:
            return self._custom_data

        @custom_data.setter
        def custom_data(self, custom_data: dict) -> None:
            self._custom_data = custom_data

        def set_shape(self, shape) -> None:
            self._shape = shape

        def update_custom_data(self, custom_data: dict) -> None:
            self.custom_data.update(custom_data)

        def update(
            self,
            item_path: str = None,
            ann_data: Union[str, dict] = None,
            shape: Union[Tuple, List] = None,
            custom_data: dict = {},
        ) -> None:
            if item_path is not None:
                self.path = item_path
            if ann_data is not None:
                self.ann_data = ann_data
            if shape is not None:
                self.set_shape(shape)
            if custom_data:
                self.update_custom_data(custom_data)

        def create_empty_annotation(self) -> Annotation:
            raise NotImplementedError()

    def __init__(
        self,
        input_data: str,
        labeling_interface: Optional[Union[LabelingInterface, str]] = LabelingInterface.DEFAULT,
        upload_as_links: bool = False,
        remote_files_map: Optional[Dict[str, str]] = None,
    ):
        self._input_data: str = input_data
        self._items: List[self.BaseItem] = []
        self._meta: ProjectMeta = None
        self._labeling_interface = labeling_interface or LabelingInterface.DEFAULT
        self._upload_as_links: bool = upload_as_links
        self._remote_files_map: Optional[Dict[str, str]] = remote_files_map
        self._supports_links = False  # if converter supports uploading by links
        self._converter = None

        if self._labeling_interface not in LabelingInterface.values():
            raise ValueError(
                f"Invalid labeling interface value: {labeling_interface}. "
                f"The available values: {LabelingInterface.values()}"
            )

    @property
    def format(self) -> str:
        return self.__str__()

    @property
    def items_count(self) -> int:
        return len(self._items)

    @property
    def ann_ext(self) -> str:
        raise NotImplementedError()

    @property
    def key_file_ext(self) -> str:
        raise NotImplementedError()

    @property
    def upload_as_links(self) -> bool:
        return self._upload_as_links

    @property
    def remote_files_map(self) -> Dict[str, str]:
        return self._remote_files_map

    @property
    def supports_links(self) -> bool:
        return self._supports_links

    def validate_labeling_interface(self) -> bool:
        return self._labeling_interface == LabelingInterface.DEFAULT

    def validate_ann_file(self, ann_path) -> bool:
        raise NotImplementedError()

    def validate_key_file(self) -> bool:
        raise NotImplementedError()

    def detect_format(self) -> BaseConverter:
        self._converter = self._detect_format()
        return self._converter

    def validate_format(self) -> bool:
        """
        Validate format of the input data meets the requirements of the converter. Should be implemented in the subclass.
        Additionally, this method must do the following steps:
            1. creates project meta (if key file file exists) and save it to self._meta
            2. creates items, count detected annotations and save them to self._items
            3. validates annotation files (and genereate meta if key file is missing)

        :return: True if format is valid, False otherwise.
        """
        raise NotImplementedError()

    def get_meta(self) -> ProjectMeta:
        return self._meta

    def get_items(self) -> List[BaseItem]:
        return self._items

    def to_supervisely(self, item: BaseItem, meta: ProjectMeta, *args) -> Annotation:
        """Convert to Supervisely format."""
        return item.create_empty_annotation()

    def upload_dataset(self, api: Api, dataset_id: int, batch_size: int = 50) -> None:
        """Upload converted data to Supervisely"""
        raise NotImplementedError()

    def _detect_format(self):
        found_formats = []
        all_converters = self.__class__.__subclasses__()

        progress, progress_cb = self.get_progress(1, "Detecting annotation format")
        for converter in all_converters:
            if converter.__name__ == "BaseConverter":
                continue
            converter = converter(
                self._input_data,
                self._labeling_interface,
                self._upload_as_links,
                self._remote_files_map,
            )

            if not converter.validate_labeling_interface():
                continue

            if self.upload_as_links and not converter.supports_links:
                continue

            if converter.validate_format():
                logger.info(f"Detected format: {str(converter)}")
                found_formats.append(converter)
                if len(found_formats) > 1:
                    raise RuntimeError(
                        f"Multiple formats detected: {[str(f) for f in found_formats]}. "
                        "Mixed formats are not supported yet."
                    )

        progress_cb(1)

        if len(found_formats) == 0:
            self._items, only_modality_items, unsupported_exts = (
                self._collect_items_if_format_not_detected()
            )

            if self.items_count == 0:
                if unsupported_exts:
                    raise RuntimeError(
                        f"Not found any {self.modality} to upload. "  # pylint: disable=no-member
                        f"Unsupported file extensions detected: {unsupported_exts}. "
                        f"Convert your data to one of the supported formats: {self.allowed_exts}"
                    )
                raise RuntimeError(
                    "Please refer to the app overview and documentation for annotation formats, "
                    "and ensure that your data contains valid information"
                )
            if not only_modality_items:
                logger.warn(
                    "Annotations not found. "  # pylint: disable=no-member
                    f"Uploading {self.modality} without annotations. "
                    "If you need assistance to upload data with annotations, please contact our support team."
                )
            return self

        if len(found_formats) == 1:
            return found_formats[0]

    def _collect_items_if_format_not_detected(self):
        only_modality_items = True
        unsupported_exts = set()
        items = []
        for root, _, files in os.walk(self._input_data):
            for file in files:
                full_path = os.path.join(root, file)
                ext = get_file_ext(full_path)
                if ext.lower() in self.allowed_exts:  # pylint: disable=no-member
                    items.append(self.Item(full_path))  # pylint: disable=no-member
                    continue
                only_modality_items = False
                if ext.lower() in self.unsupported_exts:
                    unsupported_exts.add(ext)

        return items, only_modality_items, unsupported_exts

    def merge_metas_with_conflicts(
        self, api: Api, dataset_id: int
    ) -> Tuple[ProjectMeta, dict, dict]:

        # get meta1 from project and meta2 from converter
        dataset = api.dataset.get_info_by_id(dataset_id)
        if dataset is None:
            raise RuntimeError(
                f"Dataset ID:{dataset_id} not found. "
                "Please check if the dataset exists and try again."
            )
        meta1_json = api.project.get_meta(dataset.project_id, with_settings=True)
        meta1 = ProjectMeta.from_json(meta1_json)
        meta2 = self._meta

        if meta2 is None:
            return meta1, {}, {}

        # merge classes and tags from meta1 (unchanged) and meta2 (renamed if conflict)
        renamed_classes = {}
        renamed_tags = {}
        for new_cls in meta2.obj_classes:
            i = 1
            new_name = new_cls.name
            matched = False
            while meta1.obj_classes.get(new_name) is not None:
                if meta1.obj_classes.get(new_name).geometry_type == new_cls.geometry_type:
                    matched = True
                    break
                new_name = f"{new_cls.name}_{i}"
                i += 1
            if new_name != new_cls.name:
                logger.warn(f"Class {new_cls.name} renamed to {new_name}")
                renamed_classes[new_cls.name] = new_name
            if not matched:
                new_cls = new_cls.clone(name=new_name)
                meta1 = meta1.add_obj_class(new_cls)

        for new_tag in meta2.tag_metas:
            i = 1
            new_name = new_tag.name
            matched = False
            while meta1.tag_metas.get(new_name) is not None:
                if meta1.tag_metas.get(new_name).value_type == new_tag.value_type:
                    if new_tag.value_type != TagValueType.ONEOF_STRING:
                        matched = True
                        break
                    if meta1.tag_metas.get(new_name).possible_values == new_tag.possible_values:
                        matched = True
                        break
                new_name = f"{new_tag.name}_{i}"
                i += 1
            if new_name != new_tag.name:
                logger.warn(f"Tag {new_tag.name} renamed to {new_name}")
                renamed_tags[new_tag.name] = new_name
            if not matched:
                new_tag = new_tag.clone(name=new_name)
                meta1 = meta1.add_tag_meta(new_tag)

        meta1 = self._update_labeling_interface(meta1, meta2, renamed_tags)

        # update project meta
        meta1 = api.project.update_meta(dataset.project_id, meta1)

        return meta1, renamed_classes, renamed_tags

    def get_progress(
        self,
        total: int,
        message: str = "Processing items...",
        is_size: bool = False,
    ) -> tuple:
        if is_production():
            progress = Progress(message, total, is_size=is_size)
            progress_cb = progress.iters_done_report
        else:
            progress = tqdm(
                total=total, desc=message, unit="B" if is_size else "it", unit_scale=is_size
            )
            progress_cb = progress.update
        return progress, progress_cb

    def _update_labeling_interface(
        self,
        meta1: ProjectMeta,
        meta2: ProjectMeta,
        renamed_tags: Dict[str, str] = None,
    ) -> ProjectMeta:
        """
        Update project meta with labeling interface from the converter meta.
        Only update if the existing labeling interface is the default value.
        In other cases, the existing labeling interface is preserved.
        """
        existing = meta1.project_settings.labeling_interface
        new = meta2.project_settings.labeling_interface
        if existing == new:
            return meta1

        if new is None or new == LabelingInterface.DEFAULT:
            return meta1

        if existing == LabelingInterface.DEFAULT:
            group_tag_name = meta2.project_settings.multiview_tag_name
            if group_tag_name and renamed_tags:
                group_tag_name = renamed_tags.get(group_tag_name, group_tag_name)
            new_settings = meta2.project_settings.clone(multiview_tag_name=group_tag_name)

            return meta1.clone(project_settings=new_settings)
        return meta1



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from supervisely.nn.benchmark.visualization.vis_metric_base import MetricVis
from supervisely.nn.benchmark.visualization.vis_widgets import Schema, Widget

if TYPE_CHECKING:
    from supervisely.nn.benchmark.visualization.visualizer import Visualizer


class F1ScoreAtDifferentIOU(MetricVis):

    def __init__(self, loader: Visualizer) -> None:
        super().__init__(loader)
        self.schema = Schema(
            self._loader.vis_texts,
            markdown_f1_at_ious=Widget.Markdown(
                title="Confidence Profile at Different IoU thresholds",
                is_header=True,
                formats=[self._loader.vis_texts.definitions.iou_threshold],
            ),
            chart=Widget.Chart(),
        )

    def get_figure(self, widget: Widget):  # -> Optional[go.Figure]:
        import plotly.express as px  # pylint: disable=import-error

        # score_profile = self._loader.m_full.confidence_score_profile()
        f1s = self._loader.mp.m_full.score_profile_f1s

        # downsample
        if len(self._loader.df_score_profile) > 5000:
            f1s_down = f1s[:, :: f1s.shape[1] // 1000]
        else:
            f1s_down = f1s

        iou_names = list(map(lambda x: str(round(x, 2)), self._loader.mp.iouThrs.tolist()))
        df = pd.DataFrame(
            np.concatenate([self._loader.dfsp_down["scores"].values[:, None], f1s_down.T], 1),
            columns=["scores"] + iou_names,
        )
        labels = {"value": "Value", "variable": "IoU threshold", "scores": "Confidence Score"}

        fig = px.line(
            df,
            x="scores",
            y=iou_names,
            # title="F1-Score at different IoU Thresholds",
            labels=labels,
            color_discrete_sequence=px.colors.sequential.Viridis,
            width=None,
            height=500,
        )
        fig.update_traces(
            hovertemplate="Confidence Score: %{x:.2f}<br>Value: %{y:.2f}<extra></extra>"
        )
        fig.update_layout(yaxis=dict(range=[0, 1]), xaxis=dict(range=[0, 1], tick0=0, dtick=0.1))

        # add annotations for maximum F1-Score for each IoU threshold
        for i, iou in enumerate(iou_names):
            argmax_f1 = f1s[i].argmax()
            max_f1 = f1s[i][argmax_f1]
            score = self._loader.mp.m_full.score_profile["scores"][argmax_f1]
            fig.add_annotation(
                x=score,
                y=max_f1,
                text=f"Best score: {score:.2f}",
                showarrow=True,
                arrowhead=1,
                arrowcolor="black",
                ax=0,
                ay=-30,
            )

        fig.update_layout(
            dragmode=False,
            modebar=dict(
                remove=[
                    "zoom2d",
                    "pan2d",
                    "select2d",
                    "lasso2d",
                    "zoomIn2d",
                    "zoomOut2d",
                    "autoScale2d",
                    "resetScale2d",
                ]
            ),
        )
        return fig
